###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import argparse
import yaml
import pandas as pd
import typing as t
import uuid
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2 as mlpb
from cmflib import cmfquery
from cmflib import cmf


"""
Ingest the metadata in dvc.lock file into CMF.
If cmf mlmd file exist with metadata from metrics logging or logging for other
metadata not captured in dvc.lock file, pass that mlmd file as the input file.
This code queries the file for existing pipelines, stages and executions and stores 
as a dictionary. If the execution cmd in the stored dict, matches the execution command
in the dvc.lock file, that execution is updated with additional metadata.
If there is no prior execution captured, a new execution is created
"""
uuid_ = str(uuid.uuid4())

pipeline_name = ""

parser = argparse.ArgumentParser()
"""
File name of the cmf metadata file . Pass the file name if the pipeline has been recorded
with CMF explicit log statements , to record metadata not part of dvc.lock file
"""
parser.add_argument('--cmf_filename', type=str, default="mlmd", help="cmf filename")
args = parser.parse_args()

"""
Parses the string and returns, pipeline name, context name and execution name
"""
def get_cmf_hierarchy(execution_lineage:str):
    cmf_levels = execution_lineage.split(',')
    return cmf_levels[-1], cmf_levels[1], cmf_levels[0]

"""
Ingest the metadata into cmf
args
    execution_lineage: format- execution id(if present)/execution file name/context/pipeline
      eg - demo_train.py,active_learning_training@2,active_learning -- without existing execution
         - 3,eval,active_learning -- with existing execution in cmf metadata file
    metadata: The parsed dictionary from dvc.lock file
    execution_exist : True if it exeist, False otherwise
    metawrite: cmf object 
"""
tracked = {} #Used to keep a record of files tracked by outs and therefore not needed to be tracked in deps
def ingest_metadata(execution_lineage:str, metadata:dict, metawriter:cmf.Cmf, command:str = "") :
    pipeline_name, context_name, execution = get_cmf_hierarchy(execution_lineage)

    _ = metawriter.create_execution(
        str(context_name) + '_' + str(execution), 
        {}, 
        cmd = str(command),
        create_new_execution=False
        )
        
    for k, v in metadata.items():
        if k == "deps":
            for dep in v:
                metawriter.log_dataset_with_version(dep["path"], dep["md5"], "input")
                if dep["path"] not in tracked:
                    metawriter.log_dataset(dep["path"], 'input')
        if k == "outs":
            for out in v:
                metawriter.log_dataset_with_version(out["path"], out["md5"], "output")
                tracked[out["path"]] = True

def find_location(string, elements):
    for index, element in enumerate(elements):
        if string == element:
            return index
    return None

#Query mlmd to get all the executions and its commands
cmd_exe = {}
cmf_query = cmfquery.CmfQuery(args.cmf_filename)
pipelines: t.List[str] = cmf_query.get_pipeline_names()
for pipeline in pipelines:
    pipeline_name = pipeline
    stages: t.List[str] = cmf_query.get_pipeline_stages(pipeline)
    for stage in stages:
        exe_df: pd.DataFrame = cmf_query.get_all_executions_in_stage(stage)
        """
        Parse all the executions in a stage
        eg- exe_step = ['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']
        """
        for index, row in exe_df.iterrows():
            exe_step = row['Execution']
            '''
            if already same execution command has been captured previously use the latest
            execution id to associate the new metadata
            '''
            if None is cmd_exe.get(exe_step, None):
                cmd_exe[exe_step] = str(row['id']) + "," + stage + "," + pipeline
            else:
                if row['id'] > int(cmd_exe.get(exe_step, None).split(',')[0]):
                    cmd_exe[exe_step] = str(row['id']) + "," + stage + "," + pipeline

"""
Parse the dvc.lock file.
"""
pipeline_dict = {}
with open("dvc.lock", 'r') as f:
    valuesYaml = yaml.load(f, Loader=yaml.FullLoader)

for k in valuesYaml['stages']:
    pipeline_dict[k] = {}
    commands=[]
    deps = []
    outs = []
    k_dict = {}
    i = 0
    
    for kk in valuesYaml['stages'][k]:
        if kk == 'cmd':
            cmd_list = valuesYaml['stages'][k][kk].split()
            commands.append(cmd_list)
            k_dict['cmd'] = cmd_list
            i = i + 1
        if kk == 'deps':
            deps = valuesYaml['stages'][k][kk]
            k_dict['deps'] = deps
        if kk == 'outs':
            outs = valuesYaml['stages'][k][kk]
            k_dict['outs'] = outs

    pipeline_dict[k][str(i)] = k_dict

"""
Create a unique Pipeline name if there is no mlmd file
"""


pipeline_name = "Pipeline"+"-"+str(uuid_) if not pipeline_name else pipeline_name
metawriter = cmf.Cmf(filepath = "mlmd", pipeline_name=pipeline_name, graph=True)

"""
Parse the dvc.lock dictionary and get the command section
"""
for k, v in pipeline_dict.items():
    for kk, vv in v.items():
        for kkk, vvv in vv.items():
            if kkk == 'cmd':
                """
                Key eg - cmd
                Value eg - ['python3', 'demo.py', '--enable_df', 'True']
                cmd_exe eg - {"['demo_eval.py', '--trained_model', 'data/model-3', '--enable_df', 'True', '--round', '3']":
                '3,eval,active_learning', 
                "['demo_eval.py', '--trained_model', 'data/model-2', '--enable_df', 'True', '--round', '2']": '2,eval,active_learning',
                "['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']": '1,eval,active_learning'}
                In the next line pop out the python
                if the pipeline_dict command is already there in the cmd_exe dict got from parsing the mlmd pop that cmd out 
                and use the stored lineage from the mlmd
                """
                vvv.pop(0)
                pos = find_location('--execution_name', vvv)
                if pos:
                    execution_name = vvv[pos+1]
                else:
                    execution_name = uuid_
                    
                context_name = k
                lineage = execution_name+","+context_name+","+ pipeline_name

                cmd = cmd_exe.get(str(' '.join(vvv)), None)
                _ = metawriter.create_context(pipeline_stage=context_name)

                ingest_metadata(lineage, vv, metawriter, str(' '.join(vvv)))


metawriter.log_dvc_lock("dvc.lock")



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import subprocess
from cmflib.dvc_wrapper import dvc_get_config

# This class handles pulls dvc configuration
class DvcConfig:
    @staticmethod
    def get_dvc_config():
        message = "'cmf' is not configured.\nExecute 'cmf init' command."
        result = dvc_get_config()
        if not result:
            return message
        else:
            config_list = result.split("\n")
            config_dict = {}
            for item in config_list:
                item_list = item.split("=")
                config_dict[item_list[0]] = item_list[1]
            return config_dict


def main():
    print(DvcConfig.get_dvc_config())


if __name__ == "__main__":
    main()



###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import json
import os
from cmflib import cmf_merger
from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cli.utils import find_root
from cmflib.server_interface import server_interface
from cmflib.utils.cmf_config import CmfConfig


# This class pulls mlmd file from cmf-server
class CmdMetadataPull(CmdBase):
    def run(self):
        cmfconfig = os.environ.get("CONFIG_FILE", ".cmfconfig")

        # find root_dir of .cmfconfig
        output = find_root(cmfconfig)

        # in case, there is no .cmfconfig file
        if output.find("'cmf' is  not configured") != -1:
            return output

        config_file_path = os.path.join(output, cmfconfig)
        attr_dict = CmfConfig.read_config(config_file_path)
        url = attr_dict.get("cmf-server-ip", "http://127.0.0.1:80")

        current_directory = os.getcwd()
        full_path_to_dump = ""
        data = ""
        cmd = "pull"
        mlmd_data = ""
        status = 0
        exec_id = None
        execution_flag = 0
        if self.args.file_name:  # setting directory where mlmd file will be dumped
            if not os.path.isdir(self.args.file_name):
                temp = os.path.dirname(self.args.file_name)
                if temp != "":
                    current_directory = temp
                if os.path.exists(current_directory):
                    full_path_to_dump  = self.args.file_name
                else:
                    return f"{current_directory} doesn't exists."
            else:
                return "Provide path with file name."
        else:
            full_path_to_dump = os.getcwd() + "/mlmd"
        if self.args.execution:
            exec_id = self.args.execution
        output = server_interface.call_mlmd_pull(
            url, self.args.pipeline_name, exec_id
        )  # calls cmf-server api to get mlmd file data(Json format)
        status = output.status_code
        # checks If given pipeline does not exists/ elif pull mlmd file/ else mlmd file is not available
        if output.content.decode() == None:
            return "Pipeline name " + self.args.pipeline_name + " doesn't exist."
        elif output.content.decode() == "no_exec_id":
            return f"Error: Execution id {exec_id} is not present in mlmd."
        elif output.content:
            try:
                cmf_merger.parse_json_to_mlmd(
                    output.content, full_path_to_dump, cmd, None
                )  # converts mlmd json data to mlmd file
            except Exception as e:
                return e
            # verifying status codes
            if status == 200:
                return f"SUCCESS: {full_path_to_dump} is successfully pulled."
            elif status == 404:
                return "ERROR: cmf-server is not available."
            elif status == 500:
                return "ERROR: Internal server error."
            else:
                return "ERROR: Unable to pull mlmd."
        else:
            return "mlmd file not available on cmf-server."


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Pulls mlmd from cmf-server to users's machine."

    parser = subparsers.add_parser(
        "pull",
        parents=[parent_parser],
        description="Pulls mlmd from cmf-server to users's machine.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f",
        "--file_name",
        help="Specify mlmd file name with full path.",
        metavar="<file_name>",
    )

    parser.add_argument(
        "-e", "--execution", help="Specify Execution id", metavar="<exec_id>"
    )

    parser.set_defaults(func=CmdMetadataPull)



# Copyright 2022 The Kubeflow Authors Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###


import os
import sys
import typing as t
from time import sleep
from ml_metadata.proto import metadata_store_pb2
from ml_metadata.metadata_store import metadata_store
from ipaddress import ip_address, IPv4Address
from typing import List
import functools
import uuid


def value_to_mlmd_value(value) -> metadata_store_pb2.Value:
    if value is None:
        return metadata_store_pb2.Value()
    if isinstance(value, int):
        return metadata_store_pb2.Value(int_value=value)
    if isinstance(value, float):
        return metadata_store_pb2.Value(double_value=value)
    return metadata_store_pb2.Value(string_value=str(value))


def connect_to_mlmd() -> metadata_store.MetadataStore:
    metadata_service_host = os.environ.get(
        'METADATA_GRPC_SERVICE_SERVICE_HOST', 'metadata-grpc-service')
    metadata_service_port = int(os.environ.get(
        'METADATA_GRPC_SERVICE_SERVICE_PORT', 8080))

    mlmd_connection_config = metadata_store_pb2.MetadataStoreClientConfig(
        host="[{}]".format(metadata_service_host) if isIPv6(metadata_service_host) else metadata_service_host,
        port=metadata_service_port,
    )

    # Checking the connection to the Metadata store.
    for _ in range(100):
        try:
            mlmd_store = metadata_store.MetadataStore(mlmd_connection_config)
            # All get requests fail when the DB is empty, so we have to use a put request.
            # TODO: Replace with _ = mlmd_store.get_context_types()
            # when https://github.com/google/ml-metadata/issues/28 is fixed
            _ = mlmd_store.put_execution_type(
                metadata_store_pb2.ExecutionType(
                    name="DummyExecutionType",
                )
            )
            return mlmd_store
        except Exception as e:
            print('Failed to access the Metadata store. Exception: "{}"'.format(str(e)), file=sys.stderr)
            sys.stderr.flush()
            sleep(1)

    raise RuntimeError('Could not connect to the Metadata store.')


def get_artifacts_by_id(store, artifact_id: [int]) -> List[metadata_store_pb2.Artifact]:
    try:
        artifacts = store.get_artifacts_by_id(artifact_id)
        return artifacts

    except Exception as e:
        print('Failed to get artifact. Exception: "{}"'.format(str(e)), file=sys.stderr)


def put_artifact(store, artifact: metadata_store_pb2.Artifact):
    try:
        store.put_artifacts([artifact])
    except Exception as e:
        print('Failed to put artifact . Exception: "{}"'.format(str(e)), file=sys.stderr)


def get_or_create_artifact_type(store, type_name, properties: dict = None) -> metadata_store_pb2.ArtifactType:
    try:
        artifact_type = store.get_artifact_type(type_name=type_name)
        return artifact_type
    except BaseException:
        artifact_type = metadata_store_pb2.ArtifactType(
            name=type_name,
            properties=properties,
        )
        artifact_type.id = store.put_artifact_type(artifact_type)  # Returns ID
        return artifact_type


def get_or_create_execution_type(store, type_name, properties: dict = None) -> metadata_store_pb2.ExecutionType:
    try:
        execution_type = store.get_execution_type(type_name=type_name)
        return execution_type
    except BaseException:
        execution_type = metadata_store_pb2.ExecutionType(
            name=type_name,
            properties=properties,
        )
        execution_type.id = store.put_execution_type(
            execution_type)  # Returns ID
        return execution_type


def get_or_create_context_type(store, type_name, properties: dict = None) -> metadata_store_pb2.ContextType:
    try:
        context_type = store.get_context_type(type_name=type_name)
        return context_type
    except BaseException:
        context_type = metadata_store_pb2.ContextType(
            name=type_name,
            properties=properties,
        )
        context_type.id = store.put_context_type(context_type)  # Returns ID
        return context_type


def create_artifact_with_type(
        store,
        uri: str,
        name: str,
        type_name: str,
        properties: dict = None,
        type_properties: dict = None,
        custom_properties: dict = None,
) -> metadata_store_pb2.Artifact:
    artifact_type = get_or_create_artifact_type(
        store=store,
        type_name=type_name,
        properties=type_properties,
    )
    artifact = metadata_store_pb2.Artifact(
        uri=uri,
        name=name,
        type_id=artifact_type.id,
        properties=properties,
        custom_properties=custom_properties,
    )
    artifact.id = store.put_artifacts([artifact])[0]
    return artifact


def create_execution_with_type(
        store,
        type_name: str,
        name: str,
        properties: dict = None,
        type_properties: dict = None,
        custom_properties: t.Optional[t.Dict] = None,
        create_new_execution: bool = True
) -> metadata_store_pb2.Execution:
    if create_new_execution:
        execution_type = get_or_create_execution_type(
        store = store,
        type_name = name,
        properties = type_properties,
        )
        execution = metadata_store_pb2.Execution(
            type_id=execution_type.id,
            properties=properties,
            custom_properties=custom_properties,
        )
        execution.id = store.put_executions([execution])[0]
    else:
        execution_type = get_or_create_execution_type(
        store=store,
        type_name=type_name,
        properties=type_properties,
        )
        execution = store.get_execution_by_type_and_name(type_name, name)
        if not execution:

            execution = metadata_store_pb2.Execution(
                type_id=execution_type.id,
                name=name,
                properties=properties,
                custom_properties=custom_properties,
             )
            execution.id = store.put_executions([execution])[0]
    return execution


def create_context_with_type(
        store,
        context_name: str,
        type_name: str,
        properties: dict = None,
        type_properties: dict = None,
        custom_properties: dict = None,
) -> metadata_store_pb2.Context:
    # ! Context_name must be unique
    context_type = get_or_create_context_type(
        store=store,
        type_name=type_name,
        properties=type_properties,
    )
    context = metadata_store_pb2.Context(
        name=context_name,
        type_id=context_type.id,
        properties=properties,
        custom_properties=custom_properties,
    )
    context.id = store.put_contexts([context])[0]
    return context


@functools.lru_cache(maxsize=128)
def get_context_by_name(
        store,
        context_name: str,
) -> metadata_store_pb2.Context:
    matching_contexts = [
        context for context in store.get_contexts() if context.name == context_name]
    assert len(matching_contexts) <= 1
    if len(matching_contexts) == 0:
        raise ValueError(
            'Context with name "{}" was not found'.format(context_name))
    return matching_contexts[0]


def get_or_create_context_with_type(
        store,
        context_name: str,
        type_name: str,
        properties: dict = None,
        type_properties: dict = None,
        custom_properties: dict = None,
) -> metadata_store_pb2.Context:
    try:
        context = get_context_by_name(store, context_name)
    except BaseException:
        context = create_context_with_type(
            store=store,
            context_name=context_name,
            type_name=type_name,
            properties=properties,
            type_properties=type_properties,
            custom_properties=custom_properties,
        )
    return context

    # Verifying that the context has the expected type name
    context_types = store.get_context_types_by_id([context.type_id])
    assert len(context_types) == 1
    if context_types[0].name != type_name:
        raise RuntimeError(
            'Context "{}" was found, but it has type "{}" instead of "{}"'.format(
                context_name, context_types[0].name, type_name))
    return context


def create_new_execution_in_existing_context(
        store,
        execution_type_name: str,
        execution_name: str,
        context_id: int,
        properties: dict = None,
        execution_type_properties: dict = None,
        custom_properties: dict = None,
        create_new_execution: bool = True
) -> metadata_store_pb2.Execution:
    execution = create_execution_with_type(
        store=store,
        properties=properties,
        custom_properties=custom_properties,
        type_name=execution_type_name,
        name=execution_name,
        type_properties=execution_type_properties,
        create_new_execution=create_new_execution
    )
    association = metadata_store_pb2.Association(
        execution_id=execution.id,
        context_id=context_id,
    )

    store.put_attributions_and_associations([], [association])
    return execution


# Parent context type name
PARENT_CONTEXT_TYPE_NAME = "Parent_Context"
# Project name - eg - climatenet, smartsim etc
PARENT_CONTEXT_NAME = "Pipeline"
# where is it executed - NERSC, HPE CLUSTER, VM etc
ENVIRONMENT = "Environment"
# Type of project - Active learning, Mini Epoch etc
PROJECT_TYPE = "Project_type"

# child context type name
RUN_CONTEXT_TYPE_NAME = "Child_Context"

# Pipeline stage - eg - Training, Inference, Active Labeling
PIPELINE_STAGE = "Pipeline_Stage"

# Pipeline stage - eg - Training, Inference, Active Labeling
EXECUTION_CONTEXT_NAME_PROPERTY_NAME = "Context_Type"
EXECUTION_CONTEXT_ID = "Context_ID"
EXECUTION_EXECUTION = "Execution"
EXECUTION_EXECUTION_TYPE_NAME="Execution_type_name"
EXECUTION_REPO = "Git_Repo"
EXECUTION_START_COMMIT = "Git_Start_Commit"
EXECUTION_END_COMMIT = "Git_End_Commit"
EXECUTION_PYTHON_ENV= "Python_Env"
EXECUTION_PIPELINE_TYPE = "Pipeline_Type"

EXECUTION_PIPELINE_ID = "Pipeline_id"
EXECUTION_UNIQUE_ID = "Execution_uuid"

def get_or_create_parent_context(
        store,
        pipeline: str,
        custom_properties: t.Optional[t.Dict] = None
) -> metadata_store_pb2.Context:
    mlmd_custom_properties = {}
    for property_name, property_value in (custom_properties or {}).items():
        mlmd_custom_properties[property_name] = value_to_mlmd_value(
            property_value)

    context = get_or_create_context_with_type(
        store=store,
        context_name=pipeline,
        type_name=PARENT_CONTEXT_TYPE_NAME,
        type_properties={
            PARENT_CONTEXT_NAME: metadata_store_pb2.STRING,
        },
        properties={
            PARENT_CONTEXT_NAME: metadata_store_pb2.Value(
                string_value=pipeline)},
        custom_properties=mlmd_custom_properties)
    return context


def get_or_create_run_context(
        store,
        pipeline_stage: str,
        custom_properties: t.Optional[t.Dict] = None,
) -> metadata_store_pb2.Context:
    mlmd_custom_properties = {}
    for property_name, property_value in (custom_properties or {}).items():
        mlmd_custom_properties[property_name] = value_to_mlmd_value(
            property_value)

    context = get_or_create_context_with_type(
        store=store,
        context_name=pipeline_stage,
        type_name=PIPELINE_STAGE,
        type_properties={
            PIPELINE_STAGE: metadata_store_pb2.STRING,
        },
        properties={
            PIPELINE_STAGE: metadata_store_pb2.Value(
                string_value=pipeline_stage)},
        custom_properties=mlmd_custom_properties)
    return context


def associate_child_to_parent_context(store, parent_context: metadata_store_pb2.Context,
                                      child_context: metadata_store_pb2.Context):
    try:
        associate = metadata_store_pb2.ParentContext(
            child_id=child_context.id, parent_id=parent_context.id)
        store.put_parent_contexts([associate])
    except Exception as e:
        # print(e)
        # print('Warning: Exception:{}'.format(str(e)), file=sys.stderr)
        sys.stderr.flush()


def create_new_execution_in_existing_run_context(
        store,
        execution_type_name: str = None,  # TRAINING EXECUTION
        execution_name:str = None,
        context_id: int = 0,  # TRAINING CONTEXT ASSOCIATED WITH THIS EXECUTION
        execution: str = None,
        pipeline_id: int = 0,  # THE PARENT CONTEXT
        pipeline_type: str = None,
        git_repo: str = None,
        git_start_commit: str = None,
        git_end_commit: str = "",
        python_env: str = "",
        custom_properties: {} = None,
        create_new_execution:bool = True
) -> metadata_store_pb2.Execution:
    mlmd_custom_properties = {}
    for property_name, property_value in (custom_properties or {}).items():
        mlmd_custom_properties[property_name] = value_to_mlmd_value(
            property_value)

    return create_new_execution_in_existing_context(
        store=store,
        execution_type_name=execution_type_name,
        execution_name=execution_name,
        context_id=context_id,
        execution_type_properties={
            EXECUTION_UNIQUE_ID: metadata_store_pb2.STRING,
            EXECUTION_CONTEXT_NAME_PROPERTY_NAME: metadata_store_pb2.STRING,
            EXECUTION_CONTEXT_ID: metadata_store_pb2.INT,
            EXECUTION_EXECUTION: metadata_store_pb2.STRING,
            EXECUTION_EXECUTION_TYPE_NAME: metadata_store_pb2.STRING,
            EXECUTION_PIPELINE_TYPE: metadata_store_pb2.STRING,
            EXECUTION_PIPELINE_ID: metadata_store_pb2.INT,
            EXECUTION_REPO: metadata_store_pb2.STRING,
            EXECUTION_START_COMMIT: metadata_store_pb2.STRING,
            EXECUTION_END_COMMIT: metadata_store_pb2.STRING,
            EXECUTION_PYTHON_ENV: metadata_store_pb2.STRING,
        },

        properties={

            EXECUTION_CONTEXT_NAME_PROPERTY_NAME: metadata_store_pb2.Value(string_value=execution_type_name),
            # Mistakenly used for grouping in the UX
            EXECUTION_CONTEXT_ID: metadata_store_pb2.Value(int_value=context_id),
            EXECUTION_EXECUTION: metadata_store_pb2.Value(string_value=execution),
            EXECUTION_EXECUTION_TYPE_NAME: metadata_store_pb2.Value(string_value=execution_type_name),
            EXECUTION_PIPELINE_TYPE: metadata_store_pb2.Value(string_value=pipeline_type),
            EXECUTION_PIPELINE_ID: metadata_store_pb2.Value(int_value=pipeline_id),
            EXECUTION_REPO: metadata_store_pb2.Value(string_value=git_repo),
            EXECUTION_START_COMMIT: metadata_store_pb2.Value(string_value=git_start_commit),
            EXECUTION_END_COMMIT: metadata_store_pb2.Value(string_value=git_end_commit),
            EXECUTION_PYTHON_ENV: metadata_store_pb2.Value(string_value=python_env),
            # should set to task ID, not component ID
        },
        custom_properties=mlmd_custom_properties,
        create_new_execution=create_new_execution
    )


def create_new_artifact_event_and_attribution(
        store,
        execution_id: int,
        context_id: int,
        uri: str,
        name: str,
        type_name: str,
        event_type: metadata_store_pb2.Event.Type,
        properties: dict = None,
        artifact_type_properties: dict = None,
        custom_properties: dict = None,
        artifact_name_path: metadata_store_pb2.Event.Path = None,
        milliseconds_since_epoch: int = None,
) -> metadata_store_pb2.Artifact:
    mlmd_properties = {}
    for property_name, property_value in (properties or {}).items():
        mlmd_properties[property_name] = value_to_mlmd_value(property_value)

    mlmd_custom_properties = {}
    for property_name, property_value in (custom_properties or {}).items():
        mlmd_custom_properties[property_name] = value_to_mlmd_value(
            property_value)

    artifact = create_artifact_with_type(
        store=store,
        uri=uri,
        name=name,
        type_name=type_name,
        type_properties=artifact_type_properties,
        properties=mlmd_properties,
        custom_properties=mlmd_custom_properties,
    )
    event = metadata_store_pb2.Event(
        execution_id=execution_id,
        artifact_id=artifact.id,
        type=event_type,
        path=artifact_name_path,
        milliseconds_since_epoch=milliseconds_since_epoch,
    )
    store.put_events([event])
    attribution = metadata_store_pb2.Attribution(
        context_id=context_id,
        artifact_id=artifact.id,
    )
    store.put_attributions_and_associations([attribution], [])

    return artifact


def link_execution_to_input_artifact(
        store,
        execution_id: int,
        uri: str,
        input_name: str,
) -> metadata_store_pb2.Artifact:
    artifacts = store.get_artifacts_by_uri(uri)
    if len(artifacts) == 0:
        print('Error: Not found upstream artifact with URI={}.'.format(uri), file=sys.stderr)
        return None
    if len(artifacts) > 1:
        print('Error: Found multiple artifacts with the same URI. {} Using the last one..'.format(artifacts),
              file=sys.stderr)

    artifact = artifacts[-1]

    event = metadata_store_pb2.Event(
        execution_id=execution_id,
        artifact_id=artifact.id,
        type=metadata_store_pb2.Event.INPUT,
        path=metadata_store_pb2.Event.Path(
            steps=[
                metadata_store_pb2.Event.Path.Step(
                    key=input_name,
                ),
            ]
        ),
    )
    store.put_events([event])

    return artifact


def link_execution_to_artifact(
        store,
        execution_id: int,
        uri: str,
        input_name: str,
        event_type: metadata_store_pb2.Event
) -> metadata_store_pb2.Artifact:
    artifacts = store.get_artifacts_by_uri(uri)
    if len(artifacts) == 0:
        print('Error: Not found upstream artifact with URI={}.'.format(uri), file=sys.stderr)
        return None
    if len(artifacts) > 1:
        # print('Warning: Found multiple artifacts with the same URI. {} Using the last one..'.format(artifacts),
        #      file=sys.stderr)

        print('Warning: Found multiple artifacts with the same URI.Using the last one..',
              file=sys.stderr)

    artifact = artifacts[-1]

    # Check if event already exist
    events = store.get_events_by_artifact_ids([artifact.id])
    for evt in events:
        if evt.execution_id == execution_id:
            return artifact

    event = metadata_store_pb2.Event(
        execution_id=execution_id,
        artifact_id=artifact.id,
        type=event_type,
        path=metadata_store_pb2.Event.Path(
            steps=[
                metadata_store_pb2.Event.Path.Step(
                    key=input_name,
                ),

            ]
        ),
    )
    store.put_events([event])

    return artifact


def isIPv6(ip: str) -> bool:
    try:
        return False if type(ip_address(ip)) is IPv4Address else True
    except Exception as e:
        print('Error: Exception:{}'.format(str(e)), file=sys.stderr)
        sys.stderr.flush()



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import sys


class CmfParserError(Exception):
    """Base class for CLI parser errors."""

    def __init__(self):
        super().__init__("parser error")


def parse_args(argv=None):
    """Parses CLI arguments

    Args:
        argv: optional list of arguments to parse. sys.argv is used by default.

    Raises:
        CmfParserError: raised for argument parsing errors

    """
    from .parser import get_main_parser

    parser = get_main_parser()
    args = parser.parse_args(argv)
    args.parser = parser
    return args


def main(argv=None):
    """Main entry point for cmf CLI.

    Args:
        argv: argv: optional list of arguments to parse. sys.argv is used by default.

    Returns:
        int, string: command's return code and error
    """
    args = None
    try:
        args = parse_args(argv)
        cmd = args.func(args)
        msg = cmd.do_run()
        print(msg)
    except CmfParserError:
        pass
    except KeyboardInterrupt:
        print("Interrupted by the user")
    except Exception as e:
        print(e)



###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import json
import os
from cmflib import cmf


# mlmd is created from metadata passed in Json format
def parse_json_to_mlmd(mlmd_json, path_to_store, cmd, exec_id):
    mlmd_data = json.loads(mlmd_json)
    # type(mlmd_data)
    pipelines = mlmd_data["Pipeline"]
    # print(type(pipelines))
    pipeline = pipelines[0]
    # print(type(pipeline))
    pipeline_name = pipeline["name"]
    # print(type(pipeline_name))
    stage = {}
    if cmd == "push":
        data = create_original_time_since_epoch(mlmd_data)
    else:
        data = mlmd_data
    graph = False
    if os.getenv('NEO4J_URI', "") != "":
        graph = True
    cmf_class = cmf.Cmf(filepath=path_to_store, pipeline_name=pipeline_name,
                        graph=graph, is_server=True)
    for stage in data["Pipeline"][0]["stages"]:  # Iterates over all the stages
        if exec_id is None:
            list_executions = [execution for execution in stage["executions"]]
        elif exec_id is not None:
            list_executions = [
                execution
                for execution in stage["executions"]
                if execution["id"] == int(exec_id)
            ]
        else:
            return "Invalid execution id given."

        for execution in list_executions:  # Iterates over all the executions
            _ = cmf_class.merge_created_context(
            pipeline_stage = stage['name'],
            custom_properties = stage["custom_properties"],
            )
            _ = cmf_class.merge_created_execution(
                execution["properties"]["Context_Type"],
                execution["properties"]["Execution"],
                execution["properties"],
                execution["custom_properties"],
                execution["name"]
            )
            for event in execution["events"]:  # Iterates over all the events
                artifact_type = event["artifact"]["type"]
                event_type = event["type"]
                artifact_name = (event["artifact"]["name"].split(":"))[0]
                custom_props = event["artifact"]["custom_properties"]
                props = event["artifact"]["properties"]
                # print(props,'props')
                uri = event["artifact"]["uri"]
                if artifact_type == "Dataset" and event_type == 3:
                    cmf_class.log_dataset_with_version(
                        artifact_name,
                        uri,
                        "input",
                        props,
                        custom_properties=custom_props,
                    )
                elif artifact_type == "Dataset" and event_type == 4:
                    cmf_class.log_dataset_with_version(
                        artifact_name,
                        uri,
                        "output",
                        props,
                        custom_properties=custom_props,
                    )
                elif artifact_type == "Model" and event_type == 3:
                    props["uri"] = uri
                    cmf_class.log_model_with_version(
                        path=artifact_name,
                        event="input",
                        props=props,
                        custom_properties=custom_props,
                    )
                elif artifact_type == "Model" and event_type == 4:
                    props["uri"] = uri
                    cmf_class.log_model_with_version(
                        path=artifact_name,
                        event="output",
                        props=props,
                        custom_properties=custom_props,
                    )
                elif artifact_type == "Metrics":
                    # print(props,'parse')
                    cmf_class.log_execution_metrics_from_client(event["artifact"]["name"], custom_props)
                elif artifact_type == "Dataslice":
                    dataslice = cmf_class.create_dataslice(event["artifact"]["name"])
                    dataslice.commit_existing(uri, props, custom_props)
                elif artifact_type == "Step_Metrics":
                    cmf_class.commit_existing_metrics(event["artifact"]["name"], uri, props, custom_props)
                else:
                    pass


# create_time_since_epoch is appended to mlmd pushed to cmf-server as original_create_time_since_epoch
def create_original_time_since_epoch(mlmd_data):
    stages = []
    execution = []
    artifact = []
    original_stages = []
    original_execution = []
    original_artifact = []
    mlmd_data["Pipeline"][0]["original_create_time_since_epoch"] = mlmd_data[
        "Pipeline"
    ][0]["create_time_since_epoch"]
    for i in mlmd_data["Pipeline"][0]["stages"]:
        i["custom_properties"]["original_create_time_since_epoch"] = i[
            "create_time_since_epoch"
        ]
        original_stages.append(
            i["custom_properties"]["original_create_time_since_epoch"]
        )
        stages.append(i["create_time_since_epoch"])
        # print(i['custom_properties']['original_create_time_since_epoch'])
        for j in i["executions"]:
            j["custom_properties"]["original_create_time_since_epoch"] = j[
                "create_time_since_epoch"
            ]
            original_execution.append(
                j["custom_properties"]["original_create_time_since_epoch"]
            )
            execution.append(j["create_time_since_epoch"])
            # print(j['custom_properties']['original_create_time_since_epoch'])
            for k in j["events"]:
                k["artifact"]["custom_properties"][
                    "original_create_time_since_epoch"
                ] = k["artifact"]["create_time_since_epoch"]
                original_artifact.append(
                    k["artifact"]["custom_properties"][
                        "original_create_time_since_epoch"
                    ]
                )
                artifact.append(k["artifact"]["create_time_since_epoch"])
                # print(k['artifact']['custom_properties']['original_create_time_since_epoch'])

    return mlmd_data






###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

from cmflib import cli


def _metadata_push(pipeline_name, file_name, execution_id, tensorboard):
    cli_args = cli.parse_args(
            [
               "metadata",
               "push",
               "-p",
               pipeline_name,
               "-f",
               file_name,
               "-e",
               execution_id,
               "-t",
               tensorboard
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _metadata_pull(pipeline_name, file_name, execution_id):
    cli_args = cli.parse_args(
            [
               "metadata",
               "pull",
               "-p",
               pipeline_name,
               "-f",
               file_name,
               "-e",
               execution_id,
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)

def _metadata_export(pipeline_name, json_file_name, file_name):
    cli_args = cli.parse_args(
            [
               "metadata",
               "export",
               "-p",
               pipeline_name,
               "-j",
               json_file_name,
               "-f",
               file_name,
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _artifact_push(pipeline_name, file_name):
    cli_args = cli.parse_args(
            [
               "artifact",
               "push",
               "-p",
               pipeline_name,
               "-f",
               file_name,
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


def _artifact_pull(pipeline_name, file_name):

    cli_args = cli.parse_args(
            [
               "artifact",
               "pull",
               "-p",
               pipeline_name,
               "-f",
               file_name,
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _artifact_pull_single(pipeline_name, file_name, artifact_name):
    cli_args = cli.parse_args(
            [
               "artifact",
               "pull",
               "-p",
               pipeline_name,
               "-f",
               file_name,
               "-a",
               artifact_name,
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


def _cmf_cmd_init(): 
    cli_args = cli.parse_args(
            [
               "init",
               "show"
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _init_local(path, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri):
    cli_args = cli.parse_args(
            [
               "init",
               "local",
               "--path",
               path,
               "--git-remote-url",
               git_remote_url,
               "--cmf-server-url",
               cmf_server_url,
               "--neo4j-user",
               neo4j_user,
               "--neo4j-password",
               neo4j_password,
               "--neo4j-uri",
               neo4j_uri 
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


def _init_minioS3(url, endpoint_url, access_key_id, secret_key, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri):
    cli_args = cli.parse_args(
            [
               "init",
               "minioS3",
               "--url",
               url ,  
               "--endpoint-url",
               endpoint_url,        
               "--access-key-id",
               access_key_id,
               "--secret-key",
               secret_key,
               "--git-remote-url",
               git_remote_url,
               "--cmf-server-url",
               cmf_server_url,
               "--neo4j-user",
               neo4j_user,
               "--neo4j-password",
               neo4j_password,
               "--neo4j-uri",
               neo4j_uri 
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg
    
def _init_amazonS3(url, access_key_id, secret_key, session_token, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri):
    cli_args = cli.parse_args(
            [
               "init",
               "amazonS3",
               "--url",
               url,
               "--access-key-id",
               access_key_id,
               "--secret-key",
               secret_key,
               "--session-token",
               session_token,
               "--git-remote-url",
               git_remote_url,
               "--cmf-server-url",
               cmf_server_url,
               "--neo4j-user",
               neo4j_user,
               "--neo4j-password",
               neo4j_password,
               "--neo4j-uri",
               neo4j_uri 
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _init_sshremote(path,user, port, password, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri):
    cli_args = cli.parse_args(
            [
               "init",
               "sshremote",
               "--path",
               path,
               "--user",
               user , 
               "--port",
               port,
               "--password",
               password,
               "--git-remote-url",
               git_remote_url,
               "--cmf-server-url",
               cmf_server_url,
               "--neo4j-user",
               neo4j_user,
               "--neo4j-password",
               neo4j_password,
               "--neo4j-uri",
               neo4j_uri 
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

def _init_osdfremote(path, key_id, key_path, key_issuer, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri):
    cli_args = cli.parse_args(
            [
               "init",
               "osdf",
               "--path",
               path,
               "--key-id",
               key_id, 
               "--key-path",
               key_path,
               "--key-issuer",
               key_issuer,
               "--git-remote-url",
               git_remote_url,
               "--cmf-server-url",
               cmf_server_url,
               "--neo4j-user",
               neo4j_user,
               "--neo4j-password",
               neo4j_password,
               "--neo4j-uri",
               neo4j_uri 
            ]
           )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg
    



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import requests
import json

# This function posts mlmd data to mlmd_push api on cmf-server
def call_mlmd_push(json_payload, url, exec_id, pipeline_name):
    url_to_pass = f"{url}/mlmd_push"
    json_data = {"id": exec_id, "json_payload": json_payload, "pipeline_name": pipeline_name}
    response = requests.post(url_to_pass, json=json_data)  # Post request
    # print("Status code -", response.status_code)
    return response


# This function gets mlmd data from mlmd_pull api from cmf-server
def call_mlmd_pull(url, pipeline_name, exec_id):
    url_to_pass = f"{url}/mlmd_pull/{pipeline_name}"
    response = requests.get(url_to_pass, json={"exec_id": exec_id})  # Get request
    return response


# This function posts tensorboard files to cmf-server
def call_tensorboard(url, pipeline_name, file_name, file_path):
    url_to_pass = f"{url}/tensorboard"
    files = {'file': (file_name, open(file_path, 'rb'))}
    params = {'pipeline_name': pipeline_name}
    response = requests.post(url_to_pass, files=files, params=params)
    return response



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import boto3

class AmazonS3Artifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        current_directory: str,
        bucket_name: str,
        object_name: str,
        download_loc: str,
    ):
        access_key = dvc_config_op["remote.amazons3.access_key_id"]
        secret_key = dvc_config_op["remote.amazons3.secret_access_key"]
        session_token = dvc_config_op["remote.amazons3.session_token"]
        try:
            s3 = boto3.client(
                's3',
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token
            )
            s3.head_bucket(Bucket=bucket_name)

            dir_path = ""
            if "/" in download_loc:
                dir_path, _ = download_loc.rsplit("/", 1)
            if dir_path != "":
                os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders if needed

            response = ""

            """"
            if object_name ends with .dir - it is a directory.
            we download .dir object with 'temp_dir' and remove 
            this after all the files from this .dir object is downloaded.
            """
            if object_name.endswith('.dir'):
                # in case of .dir, download_loc is a absolute path for a folder
                os.makedirs(download_loc, mode=0o777, exist_ok=True)

                # download .dir object
                temp_dir = f"{download_loc}/temp_dir"
                response = s3.download_file(bucket_name, object_name, temp_dir)

                with open(temp_dir, 'r') as file:
                    tracked_files = eval(file.read())

                # removing temp_dir
                if os.path.exists(temp_dir):
                    os.remove(temp_dir)

                """
                object_name =  files/md5/c9/d8fdacc0d942cf8d7d95b6301cfb97.dir
                contains the path of the .dir on the artifact repo
                we need to remove the hash of the .dir from the object_name
                which will leave us with the artifact repo path
                """
                repo_path = "/".join(object_name.split("/")[:-2])
                for file_info in tracked_files:
                    relpath = file_info['relpath']
                    md5_val = file_info['md5']
                    # download_loc =  /home/user/datatslice/example-get-started/test/artifacts/raw_data
                    # md5_val = a237457aa730c396e5acdbc5a64c8453
                    # we need a2/37457aa730c396e5acdbc5a64c8453
                    formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                    temp_download_loc = f"{download_loc}/{relpath}"
                    temp_object_name = f"{repo_path}/{formatted_md5}"
                    obj = s3.download_file(bucket_name, temp_object_name, temp_download_loc)
                    if obj == None:
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
            else:
                # download objects which are file
                response = s3.download_file(bucket_name, object_name, download_loc)
            if response == None:
                return f"{object_name} downloaded at {download_loc}"
            return response

        except s3.exceptions.ClientError as e:
            # If a specific error code is returned, the bucket does not exist
            if e.response['Error']['Code'] == '404':
                return f"{bucket_name}  doesn't exists!!"
            else:
                # Handle other errors
               raise
        except TypeError as exception:
            return exception
        except Exception as e:
            return e



"""This module contains all the public API for CMF"""
###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import time
import uuid
import re
import os
import sys
import pandas as pd
import typing as t

# This import is needed for jupyterlab environment
import dvc
from ml_metadata.proto import metadata_store_pb2 as mlpb
from ml_metadata.metadata_store import metadata_store
from cmflib.dvc_wrapper import (
    dvc_get_url,
    dvc_get_hash,
    git_get_commit,
    commit_output,
    git_get_repo,
    commit_dvc_lock_file,
    git_checkout_new_branch,
    check_git_repo,
    check_default_remote,
    check_git_remote,
    git_commit,
)
from cmflib import graph_wrapper
from cmflib.metadata_helper import (
    get_or_create_parent_context,
    get_or_create_run_context,
    associate_child_to_parent_context,
    create_new_execution_in_existing_run_context,
    link_execution_to_artifact,
    create_new_artifact_event_and_attribution,
    get_artifacts_by_id,
    put_artifact,
    link_execution_to_input_artifact,
)
from cmflib.utils.cmf_config import CmfConfig
from cmflib.utils.helper_functions import get_python_env, change_dir
from cmflib.cmf_commands_wrapper import (
    _metadata_push,
    _metadata_pull,
    _metadata_export,
    _artifact_pull,
    _artifact_push,
    _artifact_pull_single,
    _cmf_cmd_init,
    _init_local,
    _init_minioS3,
    _init_amazonS3,
    _init_sshremote,
    _init_osdfremote,
)

class Cmf:
    """This class provides methods to log metadata for distributed AI pipelines.
    The class instance creates an ML metadata store to store the metadata.
    It creates a driver to store nodes and its relationships to neo4j.
    The user has to provide the name of the pipeline, that needs to be recorded with CMF.
    ```python
    cmflib.cmf.Cmf(
        filepath="mlmd",
        pipeline_name="test_pipeline",
        custom_properties={"owner": "user_a"},
        graph=False
    )
    ```
    Args:
        filepath: Path  to the sqlite file to store the metadata
        pipeline_name: Name to uniquely identify the pipeline.
        Note that name is the unique identifier for a pipeline.
        If a pipeline already exist with the same name, the existing pipeline object is reused.
        custom_properties: Additional properties of the pipeline that needs to be stored.
        graph: If set to true, the libray also stores the relationships in the provided graph database.
        The following
        variables should be set: `neo4j_uri` (graph server URI), `neo4j_user` (user name) and
        `neo4j_password` (user password), e.g.:
        ```
        cmf init local --path /home/user/local-storage --git-remote-url https://github.com/XXX/exprepo.git --neo4j-user neo4j --neo4j-password neo4j
                              --neo4j-uri bolt://localhost:7687
        ```
    """

    # pylint: disable=too-many-instance-attributes
    # Reading CONFIG_FILE variable
    cmf_config = os.environ.get("CONFIG_FILE", ".cmfconfig")
    ARTIFACTS_PATH = "cmf_artifacts"
    DATASLICE_PATH = "dataslice"
    METRICS_PATH = "metrics"
    if os.path.exists(cmf_config):
        attr_dict = CmfConfig.read_config(cmf_config)
        __neo4j_uri = attr_dict.get("neo4j-uri", "")
        __neo4j_password = attr_dict.get("neo4j-password", "")
        __neo4j_user = attr_dict.get("neo4j-user", "")

    def __init__(
        self,
        filepath: str = "mlmd",
        pipeline_name: str = "",
        custom_properties: t.Optional[t.Dict] = None,
        graph: bool = False,
        is_server: bool = False,
    ):
        #path to directory
        self.cmf_init_path = filepath.rsplit("/",1)[0] \
				 if len(filepath.rsplit("/",1)) > 1 \
					else  os.getcwd()

        logging_dir = change_dir(self.cmf_init_path)
        if is_server is False:
            Cmf.__prechecks()
        if custom_properties is None:
            custom_properties = {}
        if not pipeline_name:
            # assign folder name as pipeline name 
            cur_folder = os.path.basename(os.getcwd())
            pipeline_name = cur_folder
        config = mlpb.ConnectionConfig()
        config.sqlite.filename_uri = filepath
        self.store = metadata_store.MetadataStore(config)
        self.filepath = filepath
        self.child_context = None
        self.execution = None
        self.execution_name = ""
        self.execution_command = ""
        self.metrics = {}
        self.input_artifacts = []
        self.execution_label_props = {}
        self.graph = graph
        #last token in filepath
        self.branch_name = filepath.rsplit("/", 1)[-1]

        if is_server is False:
            git_checkout_new_branch(self.branch_name)
        self.parent_context = get_or_create_parent_context(
            store=self.store,
            pipeline=pipeline_name,
            custom_properties=custom_properties,
        )
        if is_server:
            Cmf.__get_neo4j_server_config()
        if graph is True:
            Cmf.__load_neo4j_params()
            self.driver = graph_wrapper.GraphDriver(
                Cmf.__neo4j_uri, Cmf.__neo4j_user, Cmf.__neo4j_password
            )
            self.driver.create_pipeline_node(
                pipeline_name, self.parent_context.id, custom_properties
            )
        os.chdir(logging_dir)

    @staticmethod
    def __load_neo4j_params():
         cmf_config = os.environ.get("CONFIG_FILE", ".cmfconfig")
         if os.path.exists(cmf_config):
             attr_dict = CmfConfig.read_config(cmf_config)
             Cmf.__neo4j_uri = attr_dict.get("neo4j-uri", "")
             Cmf.__neo4j_password = attr_dict.get("neo4j-password", "")
             Cmf.__neo4j_user = attr_dict.get("neo4j-user", "")


    @staticmethod
    def __get_neo4j_server_config():
        Cmf.__neo4j_uri = os.getenv('NEO4J_URI', "")
        Cmf.__neo4j_user = os.getenv('NEO4J_USER_NAME', "")
        Cmf.__neo4j_password = os.getenv('NEO4J_PASSWD', "")


    @staticmethod
    def __prechecks():
        """Pre checks for cmf
        1. Needs to be a git repository and
           git remote should be set
        2. Needs to be a dvc repository and
           default dvc remote should be set
        """
        Cmf.__check_git_init()
        Cmf.__check_default_remote()
        Cmf.__check_git_remote()

    @staticmethod
    def __check_git_remote():
        """Executes precheck for git remote"""
        if not check_git_remote():
            print(
                "*** Error git remote not set ***\n"
                "*** Run cmf init ***"
            )
            sys.exit(1)

    @staticmethod
    def __check_default_remote():
        """Executes precheck for default dvc remote"""
        if not check_default_remote():
            print(
                "*** DVC not configured correctly ***\n"
                "*** Run command cmf init ***" 
            )
            sys.exit(1)

    @staticmethod
    def __check_git_init():
        """Verifies that the directory is a git repo"""
        if not check_git_repo():
            print(
                "*** Not a git repo, Please do the following ***\n"
                "*** Run Command cmf init ***"
            )
            sys.exit(1)

    def finalize(self):
        git_commit(self.execution_name)
        if self.graph:
            self.driver.close()

    def create_context(
        self, pipeline_stage: str, custom_properties: t.Optional[t.Dict] = None
    ) -> mlpb.Context:
        """Create's a  context(stage).
        Every call creates a unique pipeline stage.
        Updates Pipeline_stage name.
        Example:
            ```python
            #Create context
            # Import CMF
            from cmflib.cmf import Cmf
            from ml_metadata.proto import metadata_store_pb2 as mlpb
            # Create CMF logger
            cmf = Cmf(filepath="mlmd", pipeline_name="test_pipeline")
            # Create context
            context: mlmd.proto.Context = cmf.create_context(
                pipeline_stage="prepare",
                custom_properties ={"user-metadata1": "metadata_value"}
            )

            ```
            Args:
                Pipeline_stage: Name of the Stage.
                custom_properties: Developers can provide key value pairs with additional properties of the execution that
                    need to be stored.
            Returns:
                Context object from ML Metadata library associated with the new context for this stage.
        """
        custom_props = {} if custom_properties is None else custom_properties
        pipeline_stage = self.parent_context.name + "/" + pipeline_stage
        ctx = get_or_create_run_context(
            self.store, pipeline_stage, custom_props)
        self.child_context = ctx
        associate_child_to_parent_context(
            store=self.store, parent_context=self.parent_context, child_context=ctx
        )
        if self.graph:
            self.driver.create_stage_node(
                pipeline_stage, self.parent_context, ctx.id, custom_props
            )
        return ctx

    def merge_created_context(
        self, pipeline_stage: str, custom_properties: t.Optional[t.Dict] = None
    ) -> mlpb.Context:
        """Merge created context.
        Every call creates a unique pipeline stage.
        Created for metadata push purpose.
        Example:

            ```python
            #Create context
            # Import CMF
            from cmflib.cmf import Cmf
            from ml_metadata.proto import metadata_store_pb2 as mlpb
            # Create CMF logger
            cmf = Cmf(filepath="mlmd", pipeline_name="test_pipeline")
            # Create context
            context: mlmd.proto.Context = cmf.merge_created_context(
                pipeline_stage="Test-env/prepare",
                custom_properties ={"user-metadata1": "metadata_value"}
            ```
            Args:
                Pipeline_stage: Pipeline_Name/Stage_name.
                custom_properties: Developers can provide key value pairs with additional properties of the execution that
                    need to be stored.
            Returns:
                Context object from ML Metadata library associated with the new context for this stage.
        """

        custom_props = {} if custom_properties is None else custom_properties
        ctx = get_or_create_run_context(
            self.store, pipeline_stage, custom_props)
        self.child_context = ctx
        associate_child_to_parent_context(
            store=self.store, parent_context=self.parent_context, child_context=ctx
        )
        if self.graph:
            self.driver.create_stage_node(
                pipeline_stage, self.parent_context, ctx.id, custom_props
            )
        return ctx

    def create_execution(
        self,
        execution_type: str,
        custom_properties: t.Optional[t.Dict] = None,
        cmd: str = None,
        create_new_execution: bool = True,
    ) -> mlpb.Execution:
        """Create execution.
        Every call creates a unique execution. Execution can only be created within a context, so
        [create_context][cmflib.cmf.Cmf.create_context] must be called first.
        Example:
            ```python
            # Import CMF
            from cmflib.cmf import Cmf
            from ml_metadata.proto import metadata_store_pb2 as mlpb
            # Create CMF logger
            cmf = Cmf(filepath="mlmd", pipeline_name="test_pipeline")
            # Create or reuse context for this stage
            context: mlmd.proto.Context = cmf.create_context(
                pipeline_stage="prepare",
                custom_properties ={"user-metadata1": "metadata_value"}
            )
            # Create a new execution for this stage run
            execution: mlmd.proto.Execution = cmf.create_execution(
                execution_type="Prepare",
                custom_properties = {"split": split, "seed": seed}
            )
            ```
        Args:
            execution_type: Type of the execution.(when create_new_execution is False, this is the name of execution)
            custom_properties: Developers can provide key value pairs with additional properties of the execution that
                need to be stored.

            cmd: command used to run this execution.

            create_new_execution:bool = True, This can be used by advanced users to re-use executions
                This is applicable, when working with framework code like mmdet, pytorch lightning etc, where the
                custom call-backs are used to log metrics.
                if create_new_execution is True(Default), execution_type parameter will be used as the name of the execution type.
                if create_new_execution is False, if existing execution exist with the same name as execution_type.
                it will be reused.
                Only executions created with  create_new_execution as False will have "name" as a property.


        Returns:
            Execution object from ML Metadata library associated with the new execution for this stage.
        """
        logging_dir = change_dir(self.cmf_init_path)
        # Assigning current file name as stage and execution name
        current_script = sys.argv[0]
        file_name = os.path.basename(current_script)
        name_without_extension = os.path.splitext(file_name)[0]
        # create context if not already created
        if not self.child_context:
            self.create_context(pipeline_stage=name_without_extension)
            assert self.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

        # Initializing the execution related fields

        self.metrics = {}
        self.input_artifacts = []
        self.execution_label_props = {}
        custom_props = {} if custom_properties is None else custom_properties
        git_repo = git_get_repo()
        git_start_commit = git_get_commit()
        cmd = str(sys.argv) if cmd is None else cmd
        python_env=get_python_env()
        self.execution = create_new_execution_in_existing_run_context(
            store=self.store,
            # Type field when re-using executions
            execution_type_name=self.child_context.name,
            execution_name=execution_type, 
            #Name field if we are re-using executions
            #Type field , if creating new executions always 
            context_id=self.child_context.id,
            execution=cmd,
            pipeline_id=self.parent_context.id,
            pipeline_type=self.parent_context.name,
            git_repo=git_repo,
            git_start_commit=git_start_commit,
            python_env=python_env,
            custom_properties=custom_props,
            create_new_execution=create_new_execution,
        )
        uuids = self.execution.properties["Execution_uuid"].string_value
        if uuids:
            self.execution.properties["Execution_uuid"].string_value = uuids+","+str(uuid.uuid1())
        else:
            self.execution.properties["Execution_uuid"].string_value = str(uuid.uuid1())            
        self.store.put_executions([self.execution])
        self.execution_name = str(self.execution.id) + "," + execution_type
        self.execution_command = cmd
        for k, v in custom_props.items():
            k = re.sub("-", "_", k)
            self.execution_label_props[k] = v
        self.execution_label_props["Execution_Name"] = (
            execution_type + ":" + str(self.execution.id)
        )
        
        self.execution_label_props["execution_command"] = cmd
        if self.graph:
            self.driver.create_execution_node(
            self.execution_name,
            self.child_context.id,
            self.parent_context,
            cmd,
            self.execution.id,
            custom_props,
        )
        os.chdir(logging_dir)
        return self.execution

    def update_execution(
        self, execution_id: int, custom_properties: t.Optional[t.Dict] = None
    ):
        """Updates an existing execution.
        The custom properties can be updated after creation of the execution.
        The new custom properties is merged with earlier custom properties.
        Example
            ```python
            # Import CMF
            from cmflib.cmf import Cmf
            from ml_metadata.proto import metadata_store_pb2 as mlpb
            # Create CMF logger
            cmf = Cmf(filepath="mlmd", pipeline_name="test_pipeline")
            # Update a execution
            execution: mlmd.proto.Execution = cmf.update_execution(
                execution_id=8,
                custom_properties = {"split": split, "seed": seed}
            )
            ```
            Args:
                execution_id: id of the execution.
                custom_properties: Developers can provide key value pairs with additional properties of the execution that
                need to be updated.
            Returns:
                Execution object from ML Metadata library associated with the updated execution for this stage.
        """
        self.execution = self.store.get_executions_by_id([execution_id])[0]
        if self.execution is None:
            print("Error - no execution id")
            return
        execution_type = self.store.get_execution_types_by_id([self.execution.type_id])[
            0
        ]

        if custom_properties:
            for key, value in custom_properties.items():
                if isinstance(value, int):
                    self.execution.custom_properties[key].int_value = value
                else:
                    self.execution.custom_properties[key].string_value = str(
                        value)
        self.store.put_executions([self.execution])
        c_props = {}
        for k, v in self.execution.custom_properties.items():
            key = re.sub("-", "_", k)
            val_type = str(v).split(":", maxsplit=1)[0]
            if val_type == "string_value":
                val = self.execution.custom_properties[k].string_value
            else:
                val = str(v).split(":")[1]
            # The properties value are stored in the format type:value hence,
            # taking only value
            self.execution_label_props[key] = val
            c_props[key] = val
        self.execution_name = str(self.execution.id) + \
            "," + execution_type.name
        self.execution_command = self.execution.properties["Execution"]
        self.execution_label_props["Execution_Name"] = (
            execution_type.name + ":" + str(self.execution.id)
        )
        self.execution_label_props["execution_command"] = self.execution.properties[
            "Execution"
        ].string_value
        if self.graph:
            self.driver.create_execution_node(
                self.execution_name,
                self.child_context.id,
                self.parent_context,
                self.execution.properties["Execution"].string_value,
                self.execution.id,
                c_props,
            )
        return self.execution

    def merge_created_execution(
        self,
        execution_type: str,
        execution_cmd: str,
        properties: t.Optional[t.Dict] = None,
        custom_properties: t.Optional[t.Dict] = None,
        orig_execution_name:str = "",
        create_new_execution:bool = True
    ) -> mlpb.Execution:
        """Merge Created execution.
        Every call creates a unique execution. Execution can only be created within a context, so
        [create_context][cmflib.cmf.Cmf.create_context] must be called first.
        Every call occurs when metadata push or pull is processed. Data from pre-existing executions is used
        to create new executions with additional data(Required on cmf-server).
        Example:
            ```python
            # Import CMF
            from cmflib.cmf import Cmf
            from ml_metadata.proto import metadata_store_pb2 as mlpb
            # Create CMF logger
            cmf = Cmf(filepath="mlmd", pipeline_name="test_pipeline")
            # Create or reuse context for this stage
            context: mlmd.proto.Context = cmf.merge_created_context(
                pipeline_stage="prepare",
                custom_properties ={"user-metadata1": "metadata_value"}
            )
            # Create a new execution for this stage run
            execution: mlmd.proto.Execution = cmf.merge_created_execution(
                execution_type="Prepare",
                properties={"Context_Type":""},
                custom_properties = {"split": split, "seed": seed},
                orig_execution_name=execution_name
            )
            ```
        Args:
            execution_type: Type of the execution.(when create_new_execution is False, this is the name of execution)
            properties: Properties of Execution.
            custom_properties: Developers can provide key value pairs with additional properties of the execution that
                need to be stored.

            cmd: command used to run this execution.

            create_new_execution:bool = True, This can be used by advanced users to re-use executions
                This is applicable, when working with framework code like mmdet, pytorch lightning etc, where the
                custom call-backs are used to log metrics.
                if create_new_execution is True(Default), execution_type parameter will be used as the name of the execution type.
                if create_new_execution is False, if existing execution exist with the same name as execution_type.
                it will be reused.
                Only executions created with  create_new_execution as False will have "name" as a property.


        Returns:
            Execution object from ML Metadata library associated with the execution for this stage.
        """
        # Initializing the execution related fields
        properties = {} if properties is None else properties
        self.metrics = {}
        self.input_artifacts = []
        self.execution_label_props = {}
        custom_props = {} if custom_properties is None else custom_properties
        # print(custom_props)
        git_repo = properties.get("Git_Repo", "")
        git_start_commit = properties.get("Git_Start_Commit", "")
        python_env = properties.get("Python_Env", "")
        #name = properties.get("Name", "")
        create_new_execution = True
        execution_name = execution_type
        #exe.name property is passed as the orig_execution_name.
        #if name is not an empty string then we are re-using executions
        if orig_execution_name != "":
            create_new_execution = False
            execution_name = orig_execution_name

        self.execution = create_new_execution_in_existing_run_context(
            store=self.store,
            execution_type_name=execution_type, # Type field when re-using executions
            execution_name=execution_name, #Name field if we are re-using executionsname
                                           #Type field , if creating new executions always
            context_id=self.child_context.id,
            execution=execution_cmd,
            pipeline_id=self.parent_context.id,
            pipeline_type=self.parent_context.name,
            git_repo=git_repo,
            git_start_commit=git_start_commit,
            python_env=python_env,
            custom_properties=custom_props,
            create_new_execution=create_new_execution
        )

        uuids = ""

        uuids = self.execution.properties["Execution_uuid"].string_value
        if uuids:
            self.execution.properties["Execution_uuid"].string_value = uuids +\
                ","+properties["Execution_uuid"]
        else:
            self.execution.properties["Execution_uuid"].string_value =\
                properties["Execution_uuid"]

        
        self.store.put_executions([self.execution])
        self.execution_name = str(self.execution.id) + "," + execution_type
        self.execution_command = execution_cmd
        for k, v in custom_props.items():
            k = re.sub("-", "_", k)
            self.execution_label_props[k] = v
        self.execution_label_props["Execution_Name"] = (
            execution_type + ":" + str(self.execution.id)
        )
        self.execution_label_props["execution_command"] = execution_cmd
        if self.graph:
            self.driver.create_execution_node(
                self.execution_name,
                self.child_context.id,
                self.parent_context,
                execution_cmd,
                self.execution.id,
                custom_props,
            )
        return self.execution

    def log_dvc_lock(self, file_path: str):
        """Used to update the dvc lock file created with dvc run command."""
        print("Entered dvc lock file commit")
        return commit_dvc_lock_file(file_path, self.execution.id)

    def log_dataset(
        self,
        url: str,
        event: str,
        custom_properties: t.Optional[t.Dict] = None,
        external: bool = False,
    ) -> mlpb.Artifact:
        """Logs a dataset as artifact.
        This call adds the dataset to dvc. The dvc metadata file created (.dvc) will be added to git and committed. The
        version of the  dataset is automatically obtained from the versioning software(DVC) and tracked as a metadata.
        Example:
            ```python
            artifact: mlmd.proto.Artifact = cmf.log_dataset(
                url="/repo/data.xml",
                event="input",
                custom_properties={"source":"kaggle"}
            )
            ```
        Args:
             url: The path to the dataset.
             event: Takes arguments `INPUT` OR `OUTPUT`.
             custom_properties: Dataset properties (key/value pairs).
        Returns:
            Artifact object from ML Metadata library associated with the new dataset artifact.
        """
        logging_dir = change_dir(self.cmf_init_path)
        # Assigning current file name as stage and execution name
        current_script = sys.argv[0]
        file_name = os.path.basename(current_script)
        name_without_extension = os.path.splitext(file_name)[0]
        # create context if not already created
        if not self.child_context:
            self.create_context(pipeline_stage=name_without_extension)
            assert self.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

        # create execution if not already created
        if not self.execution:
            self.create_execution(execution_type=name_without_extension)
            assert self.execution is not None, f"Failed to create execution for {self.pipeline_name}!!"

                ### To Do : Technical Debt. 
        # If the dataset already exist , then we just link the existing dataset to the execution
        # We do not update the dataset properties . 
        # We need to append the new properties to the existing dataset properties
        custom_props = {} if custom_properties is None else custom_properties
        git_repo = git_get_repo()
        name = re.split("/", url)[-1]
        event_type = mlpb.Event.Type.OUTPUT
        existing_artifact = []
        if event.lower() == "input":
            event_type = mlpb.Event.Type.INPUT

        commit_output(url, self.execution.id)
        c_hash = dvc_get_hash(url)

        if c_hash == "":
            print("Error in getting the dvc hash,return without logging")
            return

        dataset_commit = c_hash
        dvc_url = dvc_get_url(url)
        dvc_url_with_pipeline = f"{self.parent_context.name}:{dvc_url}"
        url = url + ":" + c_hash
        if c_hash and c_hash.strip:
            existing_artifact.extend(self.store.get_artifacts_by_uri(c_hash))

        # To Do - What happens when uri is the same but names are different
        if existing_artifact and len(existing_artifact) != 0:
            existing_artifact = existing_artifact[0]

            # Quick fix- Updating only the name
            if custom_properties is not None:
                self.update_existing_artifact(
                    existing_artifact, custom_properties)
            uri = c_hash
            # update url for existing artifact
            self.update_dataset_url(existing_artifact, dvc_url_with_pipeline)
            artifact = link_execution_to_artifact(
                store=self.store,
                execution_id=self.execution.id,
                uri=uri,
                input_name=url,
                event_type=event_type,
            )
        else:
            # if((existing_artifact and len(existing_artifact )!= 0) and c_hash != ""):
            #   url = url + ":" + str(self.execution.id)
            uri = c_hash if c_hash and c_hash.strip() else str(uuid.uuid1())
            artifact = create_new_artifact_event_and_attribution(
                store=self.store,
                execution_id=self.execution.id,
                context_id=self.child_context.id,
                uri=uri,
                name=url,
                type_name="Dataset",
                event_type=event_type,
                properties={
                    "git_repo": str(git_repo),
                    # passing c_hash value to commit
                    "Commit": str(dataset_commit),
                    "url": str(dvc_url_with_pipeline),
                },
                artifact_type_properties={
                    "git_repo": mlpb.STRING,
                    "Commit": mlpb.STRING,
                    "url": mlpb.STRING,
                },
                custom_properties=custom_props,
                milliseconds_since_epoch=int(time.time() * 1000),
            )
        custom_props["git_repo"] = git_repo
        custom_props["Commit"] = dataset_commit
        self.execution_label_props["git_repo"] = git_repo
        self.execution_label_props["Commit"] = dataset_commit

        if self.graph:
            self.driver.create_dataset_node(
                name,
                url,
                uri,
                event,
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            if event.lower() == "input":
                self.input_artifacts.append(
                    {
                        "Name": name,
                        "Path": url,
                        "URI": uri,
                        "Event": event.lower(),
                        "Execution_Name": self.execution_name,
                        "Type": "Dataset",
                        "Execution_Command": self.execution_command,
                        "Pipeline_Id": self.parent_context.id,
                        "Pipeline_Name": self.parent_context.name,
                    }
                )
                self.driver.create_execution_links(uri, name, "Dataset")
            else:
                child_artifact = {
                    "Name": name,
                    "Path": url,
                    "URI": uri,
                    "Event": event.lower(),
                    "Execution_Name": self.execution_name,
                    "Type": "Dataset",
                    "Execution_Command": self.execution_command,
                    "Pipeline_Id": self.parent_context.id,
                    "Pipeline_Name": self.parent_context.name,
                }
                self.driver.create_artifact_relationships(
                    self.input_artifacts, child_artifact, self.execution_label_props
                )
        os.chdir(logging_dir)
        return artifact

    def update_dataset_url(self, artifact: mlpb.Artifact, updated_url: str):
        """Update dataset url
           Updates url of given artifact.
           Example
               ```python
               artifact: mlmd.proto.Artifact = cmf.update_dataset_url(
                artifact="data.xml.gz"
                updated_url="/repo/data.xml",
               )
               ```
               Args:
                  artifact: Artifact for which url is to be updated
                  updated_url: The updated url path of the dataset.
               Returns:
                  Updates artifact in mlmd, does not returns anything.
        """
        for key, value in artifact.properties.items():
            if key == "url":
                old_url = value.string_value
                if updated_url not in old_url:
                    new_url = f"{old_url},{updated_url}"
                    artifact.properties[key].string_value = new_url
        put_artifact(self.store, artifact)

    def update_model_url(self, dup_artifact: list, updated_url: str):
        """Updates the URL property of model artifacts.
           Example: 
               ```python
               dup_artifact = [...] # List of artifacts
               updated_url = "/new/url"
               updated_artifacts = cmf.update_model_url(dup_artifact, updated_url)
               ```
               Args:
                  dup_artifact: List of artifacts to update.
                  updated_url: New URL to add to the existing URLs.
               Returns:
                  List of updated artifacts.
        """
        for art in dup_artifact:
            dup_art = art
            for key, value in dup_art.properties.items():
                if key == "url":
                    old_url = value.string_value
                    if updated_url not in old_url:
                        new_url = f"{old_url},{updated_url}"
                        dup_art.properties[key].string_value = new_url
            put_artifact(self.store, dup_art)
        return dup_artifact

    def log_dataset_with_version(
        self,
        url: str,
        version: str,
        event: str,
        props: t.Optional[t.Dict] = None,
        custom_properties: t.Optional[t.Dict] = None,
    ) -> mlpb.Artifact:
        """Logs a dataset when the version (hash) is known.
           Example: 
             ```python 
             artifact: mlpb.Artifact = cmf.log_dataset_with_version( 
                 url="path/to/dataset", 
                 version="abcdef",
                 event="output",
                 props={ "git_repo": "https://github.com/example/repo",
                         "url": "/path/in/repo", },
                 custom_properties={ "custom_key": "custom_value", }, 
                 ) 
             ```
             Args: 
                url: Path to the dataset. 
                version: Hash or version identifier for the dataset. 
                event: Takes arguments `INPUT` or `OUTPUT`. 
                props: Optional properties for the dataset (e.g., git_repo, url). 
                custom_properties: Optional custom properties for the dataset.
             Returns:
                Artifact object from the ML Protocol Buffers library associated with the new dataset artifact. 
        """

        props = {} if props is None else props
        custom_props = {} if custom_properties is None else custom_properties
        git_repo = props.get("git_repo", "")
        name = url
        event_type = mlpb.Event.Type.OUTPUT
        existing_artifact = []
        c_hash = version
        if event.lower() == "input":
            event_type = mlpb.Event.Type.INPUT

        # dataset_commit = commit_output(url, self.execution.id)

        dataset_commit = version
        url = url + ":" + c_hash
        if c_hash and c_hash.strip:
            existing_artifact.extend(self.store.get_artifacts_by_uri(c_hash))

        # To Do - What happens when uri is the same but names are different
        if existing_artifact and len(existing_artifact) != 0:
            existing_artifact = existing_artifact[0]

            # Quick fix- Updating only the name
            if custom_properties is not None:
                self.update_existing_artifact(
                    existing_artifact, custom_properties)
            uri = c_hash
            # update url for existing artifact
            self.update_dataset_url(existing_artifact, props.get("url", ""))
            artifact = link_execution_to_artifact(
                store=self.store,
                execution_id=self.execution.id,
                uri=uri,
                input_name=url,
                event_type=event_type,
            )
        else:
            # if((existing_artifact and len(existing_artifact )!= 0) and c_hash != ""):
            #   url = url + ":" + str(self.execution.id)
            uri = c_hash if c_hash and c_hash.strip() else str(uuid.uuid1())
            artifact = create_new_artifact_event_and_attribution(
                store=self.store,
                execution_id=self.execution.id,
                context_id=self.child_context.id,
                uri=uri,
                name=url,
                type_name="Dataset",
                event_type=event_type,
                properties={
                    "git_repo": str(git_repo),
                    "Commit": str(dataset_commit),
                    "url": props.get("url", " "),
                },
                artifact_type_properties={
                    "git_repo": mlpb.STRING,
                    "Commit": mlpb.STRING,
                    "url": mlpb.STRING,
                },
                custom_properties=custom_props,
                milliseconds_since_epoch=int(time.time() * 1000),
            )
        custom_props["git_repo"] = git_repo
        custom_props["Commit"] = dataset_commit
        self.execution_label_props["git_repo"] = git_repo
        self.execution_label_props["Commit"] = dataset_commit

        if self.graph:
            self.driver.create_dataset_node(
                name,
                url,
                uri,
                event,
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            if event.lower() == "input":
                self.input_artifacts.append(
                    {
                        "Name": name,
                        "Path": url,
                        "URI": uri,
                        "Event": event.lower(),
                        "Execution_Name": self.execution_name,
                        "Type": "Dataset",
                        "Execution_Command": self.execution_command,
                        "Pipeline_Id": self.parent_context.id,
                        "Pipeline_Name": self.parent_context.name,
                    }
                )
                self.driver.create_execution_links(uri, name, "Dataset")
            else:
                child_artifact = {
                    "Name": name,
                    "Path": url,
                    "URI": uri,
                    "Event": event.lower(),
                    "Execution_Name": self.execution_name,
                    "Type": "Dataset",
                    "Execution_Command": self.execution_command,
                    "Pipeline_Id": self.parent_context.id,
                    "Pipeline_Name": self.parent_context.name,
                }
                self.driver.create_artifact_relationships(
                    self.input_artifacts, child_artifact, self.execution_label_props
                )
        return artifact

    # Add the model to dvc do a git commit and store the commit id in MLMD
    def log_model(
        self,
        path: str,
        event: str,
        model_framework: str = "Default",
        model_type: str = "Default",
        model_name: str = "Default",
        custom_properties: t.Optional[t.Dict] = None,
    ) -> mlpb.Artifact:
        """Logs a model.
        The model is added to dvc and the metadata file (.dvc) gets committed to git.
        Example:
            ```python
            artifact: mlmd.proto.Artifact= cmf.log_model(
                path="path/to/model.pkl",
                event="output",
                model_framework="SKlearn",
                model_type="RandomForestClassifier",
                model_name="RandomForestClassifier:default"
            )
            ```
        Args:
            path: Path to the model file.
            event: Takes arguments `INPUT` OR `OUTPUT`.
            model_framework: Framework used to create the model.
            model_type: Type of model algorithm used.
            model_name: Name of the algorithm used.
            custom_properties: The model properties.
        Returns:
            Artifact object from ML Metadata library associated with the new model artifact.
        """

        logging_dir = change_dir(self.cmf_init_path)
        # Assigning current file name as stage and execution name
        current_script = sys.argv[0]
        file_name = os.path.basename(current_script)
        name_without_extension = os.path.splitext(file_name)[0]
        # create context if not already created
        if not self.child_context:
            self.create_context(pipeline_stage=name_without_extension)
            assert self.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

        # create execution if not already created
        if not self.execution:
            self.create_execution(execution_type=name_without_extension)
            assert self.execution is not None, f"Failed to create execution for {self.pipeline_name}!!"


        # To Do : Technical Debt. 
        # If the model already exist , then we just link the existing model to the execution
        # We do not update the model properties . 
        # We need to append the new properties to the existing model properties
        if custom_properties is None:
            custom_properties = {}
        custom_props = {} if custom_properties is None else custom_properties
        # name = re.split('/', path)[-1]
        event_type = mlpb.Event.Type.OUTPUT
        existing_artifact = []
        if event.lower() == "input":
            event_type = mlpb.Event.Type.INPUT

        commit_output(path, self.execution.id)
        c_hash = dvc_get_hash(path)

        if c_hash == "":
            print("Error in getting the dvc hash,return without logging")
            return

        model_commit = c_hash

        # If connecting to an existing artifact - The name of the artifact is
        # used as path/steps/key
        model_uri = path + ":" + c_hash
        dvc_url = dvc_get_url(path, False)
        url = dvc_url
        url_with_pipeline = f"{self.parent_context.name}:{url}"
        uri = ""
        if c_hash and c_hash.strip():
            uri = c_hash.strip()
            existing_artifact.extend(self.store.get_artifacts_by_uri(uri))
        else:
            raise RuntimeError("Model commit failed, Model uri empty")

        if (
            existing_artifact
            and len(existing_artifact) != 0
        ):
            # update url for existing artifact
            existing_artifact = self.update_model_url(
                existing_artifact, url_with_pipeline
            )
            artifact = link_execution_to_artifact(
                store=self.store,
                execution_id=self.execution.id,
                uri=c_hash,
                input_name=model_uri,
                event_type=event_type,
            )
            model_uri = artifact.name
        else:
            uri = c_hash if c_hash and c_hash.strip() else str(uuid.uuid1())
            model_uri = model_uri + ":" + str(self.execution.id)
            artifact = create_new_artifact_event_and_attribution(
                store=self.store,
                execution_id=self.execution.id,
                context_id=self.child_context.id,
                uri=uri,
                name=model_uri,
                type_name="Model",
                event_type=event_type,
                properties={
                    "model_framework": str(model_framework),
                    "model_type": str(model_type),
                    "model_name": str(model_name),
                    # passing c_hash value to commit
                    "Commit": str(model_commit),
                    "url": str(url_with_pipeline),
                },
                artifact_type_properties={
                    "model_framework": mlpb.STRING,
                    "model_type": mlpb.STRING,
                    "model_name": mlpb.STRING,
                    "Commit": mlpb.STRING,
                    "url": mlpb.STRING,
                },
                custom_properties=custom_props,
                milliseconds_since_epoch=int(time.time() * 1000),
            )
        # custom_properties["Commit"] = model_commit
        self.execution_label_props["Commit"] = model_commit
        #To DO model nodes should be similar to dataset nodes when we create neo4j
        if self.graph:
            self.driver.create_model_node(
                model_uri,
                uri,
                event,
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            if event.lower() == "input":
                self.input_artifacts.append(
                    {
                        "Name": model_uri,
                        "URI": uri,
                        "Event": event.lower(),
                        "Execution_Name": self.execution_name,
                        "Type": "Model",
                        "Execution_Command": self.execution_command,
                        "Pipeline_Id": self.parent_context.id,
                        "Pipeline_Name": self.parent_context.name,
                    }
                )
                self.driver.create_execution_links(uri, model_uri, "Model")
            else:
                child_artifact = {
                    "Name": model_uri,
                    "URI": uri,
                    "Event": event.lower(),
                    "Execution_Name": self.execution_name,
                    "Type": "Model",
                    "Execution_Command": self.execution_command,
                    "Pipeline_Id": self.parent_context.id,
                    "Pipeline_Name": self.parent_context.name,
                }

                self.driver.create_artifact_relationships(
                    self.input_artifacts, child_artifact, self.execution_label_props
                )
        os.chdir(logging_dir)
        return artifact

    # Add the model to dvc do a git commit and store the commit id in MLMD
    def log_model_with_version(
        self,
        path: str,
        event: str,
        props=None,
        custom_properties: t.Optional[t.Dict] = None,
    ) -> object:
        """Logs a model when the version(hash) is known
         The model is added to dvc and the metadata file (.dvc) gets committed to git.
        Example:
            ```python
            artifact: mlmd.proto.Artifact= cmf.log_model_with_version(
                path="path/to/model.pkl",
                event="output",
                props={
                        "url": "/home/user/local-storage/bf/629ccd5cd008066b72c04f9a918737",
                        "model_type": "RandomForestClassifier",
                        "model_name": "RandomForestClassifier:default",
                        "Commit": "commit 1146dad8b74cae205db6a3132ea403db1e4032e5",
                        "model_framework": "SKlearn",
                       },
                custom_properties={
                        "uri": "bf629ccd5cd008066b72c04f9a918737",
                },

            )
            ```
        Args:
            path: Path to the model file.
            event: Takes arguments `INPUT` OR `OUTPUT`.
            props: Model artifact properties.
            custom_properties: The model properties.
        Returns:
            Artifact object from ML Metadata library associated with the new model artifact.
        """

        if custom_properties is None:
            custom_properties = {}
        custom_props = {} if custom_properties is None else custom_properties
        name = re.split("/", path)[-1]
        event_type = mlpb.Event.Type.OUTPUT
        existing_artifact = []
        if event.lower() == "input":
            event_type = mlpb.Event.Type.INPUT

        # props["commit"] = "" # To do get from incoming data
        c_hash = props.get("uri", " ")
        # If connecting to an existing artifact - The name of the artifact is used as path/steps/key
        model_uri = path + ":" + c_hash
        # dvc_url = dvc_get_url(path, False)
        url = props.get("url", "")
        # uri = ""
        if c_hash and c_hash.strip():
            uri = c_hash.strip()
            existing_artifact.extend(self.store.get_artifacts_by_uri(uri))
        else:
            raise RuntimeError("Model commit failed, Model uri empty")

        if (
            existing_artifact
            and len(existing_artifact) != 0
        ):
            # update url for existing artifact
            existing_artifact = self.update_model_url(existing_artifact, url)
            artifact = link_execution_to_artifact(
                store=self.store,
                execution_id=self.execution.id,
                uri=c_hash,
                input_name=model_uri,
                event_type=event_type,
            )
            model_uri = artifact.name
        else:
            uri = c_hash if c_hash and c_hash.strip() else str(uuid.uuid1())
            model_uri = model_uri + ":" + str(self.execution.id)
            artifact = create_new_artifact_event_and_attribution(
                store=self.store,
                execution_id=self.execution.id,
                context_id=self.child_context.id,
                uri=uri,
                name=model_uri,
                type_name="Model",
                event_type=event_type,
                properties={
                    "model_framework": props.get("model_framework", ""),
                    "model_type": props.get("model_type", ""),
                    "model_name": props.get("model_name", ""),
                    "Commit": props.get("Commit", ""),
                    "url": str(url),
                },
                artifact_type_properties={
                    "model_framework": mlpb.STRING,
                    "model_type": mlpb.STRING,
                    "model_name": mlpb.STRING,
                    "Commit": mlpb.STRING,
                    "url": mlpb.STRING,
                },
                custom_properties=custom_props,
                milliseconds_since_epoch=int(time.time() * 1000),
            )
        # custom_properties["Commit"] = model_commit
        # custom_props["url"] = url
        self.execution_label_props["Commit"] = props.get("Commit", "")
        if self.graph:
            self.driver.create_model_node(
                model_uri,
                uri,
                event,
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            if event.lower() == "input":
                self.input_artifacts.append(
                    {
                        "Name": model_uri,
                        "URI": uri,
                        "Event": event.lower(),
                        "Execution_Name": self.execution_name,
                        "Type": "Model",
                        "Execution_Command": self.execution_command,
                        "Pipeline_Id": self.parent_context.id,
                        "Pipeline_Name": self.parent_context.name,
                    }
                )
                self.driver.create_execution_links(uri, model_uri, "Model")
            else:
                child_artifact = {
                    "Name": model_uri,
                    "URI": uri,
                    "Event": event.lower(),
                    "Execution_Name": self.execution_name,
                    "Type": "Model",
                    "Execution_Command": self.execution_command,
                    "Pipeline_Id": self.parent_context.id,
                    "Pipeline_Name": self.parent_context.name,
                }
                self.driver.create_artifact_relationships(
                    self.input_artifacts, child_artifact, self.execution_label_props
                )

        return artifact

    def log_execution_metrics_from_client(self, metrics_name: str,
                                         custom_properties: t.Optional[t.Dict] = None) -> mlpb.Artifact:
        """ Logs execution metrics from a client.
            Data from pre-existing metrics from client side is used to create identical metrics on server side. 
            Example: 
              ```python 
              artifact: mlpb.Artifact = cmf.log_execution_metrics_from_client( 
                      metrics_name="example_metrics:uri:123", 
                      custom_properties={"custom_key": "custom_value"}, 
                      )
              ``` 
              Args: 
                 metrics_name: Name of the metrics in the format "name:uri:execution_id". 
                 custom_properties: Optional custom properties for the metrics. 
              Returns: 
                 Artifact object from the ML Protocol Buffers library associated with the metrics artifact.
        """

        metrics = None
        custom_props = {} if custom_properties is None else custom_properties
        existing_artifact = []
        name_tokens = metrics_name.split(":")
        if name_tokens and len(name_tokens) > 2:
            name = name_tokens[0]
            uri = name_tokens[1]
            execution_id = name_tokens[2]
        else:
            print(f"Error : metrics name {metrics_name} is not in the correct format")
            return 

        #we need to add the execution id to the metrics name
        new_metrics_name = f"{name}:{uri}:{str(self.execution.id)}"
        existing_artifacts = self.store.get_artifacts_by_uri(uri)

        existing_artifact = existing_artifacts[0] if existing_artifacts else None
        if not existing_artifact or \
           ((existing_artifact) and not
            (existing_artifact.name == new_metrics_name)):  #we need to add the artifact otherwise its already there 
            metrics = create_new_artifact_event_and_attribution(
            store=self.store,
            execution_id=self.execution.id,
            context_id=self.child_context.id,
            uri=uri,
            name=new_metrics_name,
            type_name="Metrics",
            event_type=mlpb.Event.Type.OUTPUT,
            properties={"metrics_name": metrics_name},
            artifact_type_properties={"metrics_name": mlpb.STRING},
            custom_properties=custom_props,
            milliseconds_since_epoch=int(time.time() * 1000),
        )
            if self.graph:
                # To do create execution_links
                self.driver.create_metrics_node(
                    metrics_name,
                    uri,
                    "output",
                    self.execution.id,
                    self.parent_context,
                    custom_props,
                )
                child_artifact = {
                    "Name": metrics_name,
                    "URI": uri,
                    "Event": "output",
                    "Execution_Name": self.execution_name,
                    "Type": "Metrics",
                    "Execution_Command": self.execution_command,
                    "Pipeline_Id": self.parent_context.id,
                    "Pipeline_Name": self.parent_context.name,
                }
                self.driver.create_artifact_relationships(
                    self.input_artifacts, child_artifact, self.execution_label_props
                )
        return metrics


    def log_execution_metrics(
        self, metrics_name: str, custom_properties: t.Optional[t.Dict] = None
    ) -> mlpb.Artifact:
        """Log the metadata associated with the execution (coarse-grained tracking).
        It is stored as a metrics artifact. This does not have a backing physical file, unlike other artifacts that we
        have.
        Example:
            ```python
            exec_metrics: mlpb.Artifact = cmf.log_execution_metrics(
                metrics_name="Training_Metrics",
                {"auc": auc, "loss": loss}
            )
            ```
        Args:
            metrics_name: Name to identify the metrics.
            custom_properties: Dictionary with metric values.
        Returns:
              Artifact object from ML Metadata library associated with the new coarse-grained metrics artifact.
        """
        logging_dir = change_dir(self.cmf_init_path)
        # Assigning current file name as stage and execution name
        current_script = sys.argv[0]
        file_name = os.path.basename(current_script)
        name_without_extension = os.path.splitext(file_name)[0]
        # create context if not already created
        if not self.child_context:
            self.create_context(pipeline_stage=name_without_extension)
            assert self.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

        # create execution if not already created
        if not self.execution:
            self.create_execution(execution_type=name_without_extension)
            assert self.execution is not None, f"Failed to create execution for {self.pipeline_name}!!"

        custom_props = {} if custom_properties is None else custom_properties
        uri = str(uuid.uuid1())
        metrics_name = metrics_name + ":" + uri + ":" + str(self.execution.id)
        metrics = create_new_artifact_event_and_attribution(
            store=self.store,
            execution_id=self.execution.id,
            context_id=self.child_context.id,
            uri=uri,
            name=metrics_name,
            type_name="Metrics",
            event_type=mlpb.Event.Type.OUTPUT,
            properties={"metrics_name": metrics_name},
            artifact_type_properties={"metrics_name": mlpb.STRING},
            custom_properties=custom_props,
            milliseconds_since_epoch=int(time.time() * 1000),
        )
        if self.graph:
            # To do create execution_links
            self.driver.create_metrics_node(
                metrics_name,
                uri,
                "output",
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            child_artifact = {
                "Name": metrics_name,
                "URI": uri,
                "Event": "output",
                "Execution_Name": self.execution_name,
                "Type": "Metrics",
                "Execution_Command": self.execution_command,
                "Pipeline_Id": self.parent_context.id,
                "Pipeline_Name": self.parent_context.name,
            }
            self.driver.create_artifact_relationships(
                self.input_artifacts, child_artifact, self.execution_label_props
            )
        os.chdir(logging_dir)
        return metrics

    def log_metric(
        self, metrics_name: str, custom_properties: t.Optional[t.Dict] = None
    ) -> None:
        """Stores the fine-grained (per step or per epoch) metrics to memory.
        The metrics provided are stored in a parquet file. The `commit_metrics` call add the parquet file in the version
        control framework. The metrics written in the parquet file can be retrieved using the `read_metrics` call.
        Example:
            ```python
            # Can be called at every epoch or every step in the training. This is logged to a parquet file and committed
            # at the commit stage.
            # Inside training loop
            while True:
                 cmf.log_metric("training_metrics", {"train_loss": train_loss})
            cmf.commit_metrics("training_metrics")
            ```
        Args:
            metrics_name: Name to identify the metrics.
            custom_properties: Dictionary with metrics.
        """
        if metrics_name in self.metrics:
            key = max((self.metrics[metrics_name]).keys()) + 1
            self.metrics[metrics_name][key] = custom_properties
        else:
            self.metrics[metrics_name] = {}
            self.metrics[metrics_name][1] = custom_properties

    def commit_metrics(self, metrics_name: str):
        """ Writes the in-memory metrics to a Parquet file, commits the metrics file associated with the metrics id to DVC and Git,
        and stores the artifact in MLMD.

        Example:
        ```python
        artifact: mlpb.Artifact = cmf.commit_metrics("example_metrics")
        ```

        Args:
           metrics_name: Name of the metrics.

        Returns:
           Artifact object from the ML Protocol Buffers library associated with the new metrics artifact.
        """

        logging_dir = change_dir(self.cmf_init_path)
        # code for nano cmf
        # Assigning current file name as stage and execution name
        current_script = sys.argv[0]
        file_name = os.path.basename(current_script)
        name_without_extension = os.path.splitext(file_name)[0]
        # create context if not already created
        if not self.child_context:
            self.create_context(pipeline_stage=name_without_extension)
            assert self.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

        # create execution if not already created
        if not self.execution:
            self.create_execution(execution_type=name_without_extension)
            assert self.execution is not None, f"Failed to create execution for {self.pipeline_name}!!"

        
        directory_path = os.path.join(self.ARTIFACTS_PATH, self.execution.properties["Execution_uuid"].string_value.split(',')[0], self.METRICS_PATH)
        os.makedirs(directory_path, exist_ok=True)
        metrics_df = pd.DataFrame.from_dict(
            self.metrics[metrics_name], orient="index")
        metrics_df.index.names = ["SequenceNumber"]
        metrics_path = os.path.join(directory_path,metrics_name)
        metrics_df.to_parquet(metrics_path)
        commit_output(metrics_path, self.execution.id)
        uri = dvc_get_hash(metrics_path)

        if uri == "":
            print("Error in getting the dvc hash,return without logging")
            return
        metrics_commit = uri
        dvc_url = dvc_get_url(metrics_path)
        dvc_url_with_pipeline = f"{self.parent_context.name}:{dvc_url}"
        name = (
            metrics_path
            + ":"
            + uri
            + ":"
            + str(self.execution.id)
            + ":"
            + str(uuid.uuid1())
        )
        # not needed as property 'name' is part of artifact 
        # to maintain uniformity - Commit goes propeties of the artifact
        # custom_props = {"Name": metrics_name, "Commit": metrics_commit}
        custom_props = {}
        metrics = create_new_artifact_event_and_attribution(
            store=self.store,
            execution_id=self.execution.id,
            context_id=self.child_context.id,
            uri=uri,
            name=name,
            type_name="Step_Metrics",
            event_type=mlpb.Event.Type.OUTPUT,
            properties={
                # passing uri value to commit
                "Commit": metrics_commit,
                "url": str(dvc_url_with_pipeline),
            },
            artifact_type_properties={
                "Commit": mlpb.STRING,
                "url": mlpb.STRING,
            },
            custom_properties=custom_props,
            milliseconds_since_epoch=int(time.time() * 1000),
        )
        if self.graph:
            self.driver.create_metrics_node(
                name,
                uri,
                "output",
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            child_artifact = {
                "Name": name,
                "URI": uri,
                "Event": "output",
                "Execution_Name": self.execution_name,
                "Type": "Metrics",
                "Execution_Command": self.execution_command,
                "Pipeline_Id": self.parent_context.id,
            }
            self.driver.create_artifact_relationships(
                self.input_artifacts, child_artifact, self.execution_label_props
            )

        os.chdir(logging_dir)
        return metrics

    def commit_existing_metrics(self, metrics_name: str, uri: str, props: t.Optional[t.Dict] = None, custom_properties: t.Optional[t.Dict] = None):
        """
        Commits existing metrics associated with the given URI to MLMD.
        Example:
        ```python
           artifact: mlpb.Artifact = cmf.commit_existing_metrics("existing_metrics", "abc123",
           {"custom_key": "custom_value"})
        ```
        Args:
           metrics_name: Name of the metrics.
           uri: Unique identifier associated with the metrics.
           custom_properties: Optional custom properties for the metrics.
        Returns:
           Artifact object from the ML Protocol Buffers library associated with the existing metrics artifact.
        """

        custom_props =  {} if custom_properties is None else custom_properties
        c_hash = uri.strip()
        existing_artifact = []
        existing_artifact.extend(self.store.get_artifacts_by_uri(c_hash))
        if (existing_artifact
            and len(existing_artifact) != 0 ):
            metrics = link_execution_to_artifact(
                store=self.store,
                execution_id=self.execution.id,
                uri=c_hash,
                input_name=metrics_name,
                event_type=mlpb.Event.Type.OUTPUT,
            )
        else:
            metrics = create_new_artifact_event_and_attribution(
                store=self.store,
                execution_id=self.execution.id,
                context_id=self.child_context.id,
                uri=uri,
                name=metrics_name,
                type_name="Step_Metrics",
                event_type=mlpb.Event.Type.OUTPUT,
                properties={
                    # passing uri value to commit
                    "Commit": props.get("Commit", ""),
                    "url": props.get("url", ""),
                },
                artifact_type_properties={
                    "Commit": mlpb.STRING,
                    "url": mlpb.STRING,
                },
                custom_properties=custom_props,
                milliseconds_since_epoch=int(time.time() * 1000),
            )
        if self.graph:
            self.driver.create_metrics_node(
                metrics_name,
                uri,
                "output",
                self.execution.id,
                self.parent_context,
                custom_props,
            )
            child_artifact = {
                "Name": metrics_name,
                "URI": uri,
                "Event": "output",
                "Execution_Name": self.execution_name,
                "Type": "Metrics",
                "Execution_Command": self.execution_command,
                "Pipeline_Id": self.parent_context.id,
            }
            self.driver.create_artifact_relationships(
                self.input_artifacts, child_artifact, self.execution_label_props
            )
        return metrics


    def log_validation_output(
        self, version: str, custom_properties: t.Optional[t.Dict] = None
    ) -> object: 
        uri = str(uuid.uuid1())
        return create_new_artifact_event_and_attribution(
            store=self.store,
            execution_id=self.execution.id,
            context_id=self.child_context.id,
            uri=uri,
            name=uri,
            type_name="Validation_output",
            event_type=mlpb.Event.Type.INTERNAL_OUTPUT,
            properties={"version": version},
            artifact_type_properties={"version": mlpb.STRING},
            custom_properties=custom_properties,
            milliseconds_since_epoch=int(time.time() * 1000),
        )


    def update_existing_artifact(
        self, artifact: mlpb.Artifact, custom_properties: t.Dict
    ):
        """ Updates an existing artifact with the provided custom properties and stores it back to MLMD. 
          Example: 
          ```python
                update_artifact=cmf.update_existing_artifact(existing_artifact, {"key1": "updated_value"}) 
          ``` 
          Args: 
             artifact: Existing artifact to be updated. 
             custom_properties: Dictionary containing custom properties to update. 
          Returns: 
             None 
        """

        for key, value in custom_properties.items():
            if isinstance(value, int):
                artifact.custom_properties[key].int_value = value
            else:
                artifact.custom_properties[key].string_value = str(value)
        put_artifact(self.store, artifact)

    def get_artifact(self, artifact_id: int) -> mlpb.Artifact:
        """Gets the artifact object from mlmd"""
        return get_artifacts_by_id(self.store, [artifact_id])[0]

    # To Do - The interface should be simplified.
    # To Do - Links should be created in mlmd also.
    # Todo - assumes source as Dataset and target as slice - should be generic and accomodate any types
    def link_artifacts(
        self, artifact_source: mlpb.Artifact, artifact_target: mlpb.Artifact
    ):
        self.driver.create_links(artifact_source.name,
                                 artifact_target.name, "derived")

    def update_model_output(self, artifact: mlpb.Artifact):
        """updates an artifact"""
        put_artifact(self.store, artifact)

    def create_dataslice(self, name: str) -> "Cmf.DataSlice":
        """Creates a dataslice object.
        Once created, users can add data instances to this data slice with [add_data][cmflib.cmf.Cmf.DataSlice.add_data]
        method. Users are also responsible for committing data slices by calling the
        [commit][cmflib.cmf.Cmf.DataSlice.commit] method.
        Example:
            ```python
            dataslice = cmf.create_dataslice("slice-a")
            ```
        Args:
            name: Name to identify the dataslice.

        Returns:
            Instance of a newly created [DataSlice][cmflib.cmf.Cmf.DataSlice].
        """
        return Cmf.DataSlice(name, self)

    def read_dataslice(self, name: str) -> pd.DataFrame:
        """Reads the dataslice"""
        # To do checkout if not there
        directory_path = os.path.join(self.ARTIFACTS_PATH, self.execution.properties["Execution_uuid"].string_value.split(',')[0], self.DATASLICE_PATH)
        name = os.path.join(directory_path, name)
        df = pd.read_parquet(name)
        return df

    # To do - Once update the hash and the new version should be updated in
    # the mlmd
    def update_dataslice(self, name: str, record: str, custom_properties: t.Dict):
        """Updates a dataslice record in a Parquet file with the provided custom properties.
        Example:
        ```python
           dataslice=cmf.update_dataslice("dataslice_file.parquet", "record_id", 
           {"key1": "updated_value"})
        ```
        Args:
           name: Name of the Parquet file.
           record: Identifier of the dataslice record to be updated.
           custom_properties: Dictionary containing custom properties to update.

        Returns:
           None
        """
        directory_path = os.path.join(self.ARTIFACTS_PATH, self.execution.properties["Execution_uuid"].string_value.split(',')[0], self.DATASLICE_PATH)
        name = os.path.join(directory_path, name)
        df = pd.read_parquet(name)
        temp_dict = df.to_dict("index")
        temp_dict[record].update(custom_properties)
        dataslice_df = pd.DataFrame.from_dict(temp_dict, orient="index")
        dataslice_df.index.names = ["Path"]
        dataslice_df.to_parquet(name)

    class DataSlice:
        """A data slice represents a named subset of data.
        It can be used to track performance of an ML model on different slices of the training or testing dataset
        splits. This can be useful from different perspectives, for instance, to mitigate model bias.
        > Instances of data slices are not meant to be created manually by users. Instead, use
        [Cmf.create_dataslice][cmflib.cmf.Cmf.create_dataslice] method.
        """

        def __init__(self, name: str, writer):
            self.props = {}
            self.name = name
            self.writer = writer

        # def add_files(self, list_of_files:np.array ):
        #    for i in list_of_files:

        def add_data(
            self, path: str, custom_properties: t.Optional[t.Dict] = None
        ) -> None:
            """Add data to create the dataslice.
            Currently supported only for file abstractions. Pre-condition - the parent folder, containing the file
                should already be versioned.
            Example:
                ```python
                dataslice.add_data(f"data/raw_data/{j}.xml)
                ```
            Args:
                path: Name to identify the file to be added to the dataslice.
                custom_properties: Properties associated with this datum.
            """

            self.props[path] = {}
            self.props[path]['hash'] = dvc_get_hash(path)
            parent_path = path.rsplit("/", 1)[0]
            self.data_parent = parent_path.rsplit("/", 1)[1]
            if custom_properties:
                for k, v in custom_properties.items():
                    self.props[path][k] = v

        #        """
        #        Place holder for updating back to mlmd

        #        def update_data(self, path, custom_props:{}):
        #            for k ,v in custom_props.items():
        #                self.props[path][k] = v
        #        """

        def commit(self, custom_properties: t.Optional[t.Dict] = None) -> None:
            """Commit the dataslice.
            The created dataslice is versioned and added to underneath data versioning software.
            Example:

                dataslice.commit()
                ```
            Args:
                custom_properties: Dictionary to store key value pairs associated with Dataslice
                Example{"mean":2.5, "median":2.6}
            """

            logging_dir = change_dir(self.writer.cmf_init_path)
            # code for nano cmf
            # Assigning current file name as stage and execution name
            current_script = sys.argv[0]
            file_name = os.path.basename(current_script)
            name_without_extension = os.path.splitext(file_name)[0]
            # create context if not already created
            if not self.writer.child_context:
                self.writer.create_context(pipeline_stage=name_without_extension)
                assert self.writer.child_context is not None, f"Failed to create context for {self.pipeline_name}!!"

            # create execution if not already created
            if not self.writer.execution:
                self.writer.create_execution(execution_type=name_without_extension)
                assert self.writer.execution is not None, f"Failed to create execution for {self.pipeline_name}!!"

            directory_path = os.path.join(self.writer.ARTIFACTS_PATH, self.writer.execution.properties["Execution_uuid"].string_value.split(',')[0], self.writer.DATASLICE_PATH)
            os.makedirs(directory_path, exist_ok=True)
            custom_props = {} if custom_properties is None else custom_properties
            git_repo = git_get_repo()
            dataslice_df = pd.DataFrame.from_dict(self.props, orient="index")
            dataslice_df.index.names = ["Path"]
            dataslice_path = os.path.join(directory_path,self.name)
            dataslice_df.to_parquet(dataslice_path)
            existing_artifact = []

            commit_output(dataslice_path, self.writer.execution.id)
            c_hash = dvc_get_hash(dataslice_path)
            if c_hash == "":
                print("Error in getting the dvc hash,return without logging")
                return

            dataslice_commit = c_hash
            url = dvc_get_url(dataslice_path)
            dvc_url_with_pipeline = f"{self.writer.parent_context.name}:{url}"
            if c_hash and c_hash.strip():
                existing_artifact.extend(
                    self.writer.store.get_artifacts_by_uri(c_hash))
            if existing_artifact and len(existing_artifact) != 0:
                print("Adding to existing data slice")
                # Haven't added event type in this if cond, is it not needed??
                slice = link_execution_to_input_artifact(
                    store=self.writer.store,
                    execution_id=self.writer.execution.id,
                    uri=c_hash,
                    input_name=dataslice_path + ":" + c_hash,
                )
            else:
                props={
                        "git_repo": str(git_repo),
                        # passing c_hash value to commit
                        "Commit": str(dataslice_commit),
                        "url": str(dvc_url_with_pipeline),
                    },
                slice = create_new_artifact_event_and_attribution(
                    store=self.writer.store,
                    execution_id=self.writer.execution.id,
                    context_id=self.writer.child_context.id,
                    uri=c_hash,
                    name=dataslice_path + ":" + c_hash,
                    type_name="Dataslice",
                    event_type=mlpb.Event.Type.OUTPUT,
                    properties={
                        "git_repo": str(git_repo),
                        # passing c_hash value to commit
                        "Commit": str(dataslice_commit),
                        "url": str(dvc_url_with_pipeline),
                    },
                    artifact_type_properties={
                        "git_repo": mlpb.STRING,
                        "Commit": mlpb.STRING,
                        "url": mlpb.STRING,
                    },
                    custom_properties=custom_props,
                    milliseconds_since_epoch=int(time.time() * 1000),
                )
            if self.writer.graph:
                self.writer.driver.create_dataslice_node(
                    self.name, dataslice_path + ":" + c_hash, c_hash, self.data_parent, props
                )
            os.chdir(logging_dir)
            return slice

        # commit existing dataslice to server
        def commit_existing(self, uri: str, props: t.Optional[t.Dict] = None, custom_properties: t.Optional[t.Dict] = None) -> None:
            custom_props = {} if custom_properties is None else custom_properties
            c_hash = uri.strip()
            dataslice_commit = c_hash
            existing_artifact = []
            if c_hash and c_hash.strip():
                existing_artifact.extend(
                    self.writer.store.get_artifacts_by_uri(c_hash))
            if existing_artifact and len(existing_artifact) != 0:
                print("Adding to existing data slice")
                 # Haven't added event type in this if cond, is it not needed??
                slice = link_execution_to_input_artifact(
                    store=self.writer.store,
                    execution_id=self.writer.execution.id,
                    uri=c_hash,
                    input_name=self.name,
                )
            else:
                slice = create_new_artifact_event_and_attribution(
                    store=self.writer.store,
                    execution_id=self.writer.execution.id,
                    context_id=self.writer.child_context.id,
                    uri=c_hash,
                    name=self.name,
                    type_name="Dataslice",
                    event_type=mlpb.Event.Type.OUTPUT,
                    properties={
                        "git_repo": props.get("git_repo", ""),
                        "Commit": props.get("Commit", ""),
                        "url": props.get("url", " "),
                    },
                    artifact_type_properties={
                        "git_repo": mlpb.STRING,
                        "Commit": mlpb.STRING,
                        "url": mlpb.STRING,
                    },
                    custom_properties=custom_properties,
                    milliseconds_since_epoch=int(time.time() * 1000),
                )
            if self.writer.graph:
                self.writer.driver.create_dataslice_node(
                    self.name, self.name, c_hash, self.data_parent, custom_properties
                )
            return slice


#        """Temporary code"""
#
#        def materialize(self, name):
#            slicedir = name + "-" + "dir"
#            os.mkdir(slicedir)
#            df = pd.read_parquet(name)
#            for index, row in df.iterrows():
#                print(index)
#                first, middle, last = str(index).split("/")
#                print(last)
#                os.symlink(str(index), slicedir + "/ " + last)

def metadata_push(pipeline_name: str, filepath = "./mlmd", tensorboard_path: str = "", execution_id: str = ""):
    """ Pushes MLMD file to CMF-server.
    Example:
    ```python
         result = metadata_push("example_pipeline", "mlmd_file", "3")
    ```
    Args:
        pipeline_name: Name of the pipeline.
        filepath: Path to the MLMD file.
        execution_id: Optional execution ID.
        tensorboard_path: Path to tensorboard logs.

    Returns:
        Response output from the _metadata_push function.
    """
    # Required arguments:  pipeline_name
    # Optional arguments: Execution_ID, filepath (mlmd file path, tensorboard_path
    output = _metadata_push(pipeline_name, filepath, execution_id, tensorboard_path)
    return output

def metadata_pull(pipeline_name: str, filepath = "./mlmd", execution_id: str = ""):
    """ Pulls MLMD file from CMF-server. 
     Example: 
     ```python 
          result = metadata_pull("example_pipeline", "./mlmd_directory", "execution_123") 
     ``` 
     Args: 
        pipeline_name: Name of the pipeline. 
        filepath: File path to store the MLMD file. 
        execution_id: Optional execution ID. 
     Returns: 
        Message from the _metadata_pull function. 
     """
    # Required arguments:  pipeline_name 
    #Optional arguments: Execution_ID, filepath(file path to store mlmd file) 
    output = _metadata_pull(pipeline_name, filepath, execution_id)
    return output

def metadata_export(pipeline_name: str, jsonfilepath: str = "", filepath = "./mlmd"):
    """ Export local mlmd's metadata in json format to a json file. 
     Example: 
     ```python 
          result = metadata_pull("example_pipeline", "./jsonfile", "./mlmd_directory") 
     ``` 
     Args: 
        pipeline_name: Name of the pipeline. 
        jsonfilepath: File path of json file. 
        filepath: File path to store the MLMD file. 
     Returns: 
        Message from the _metadata_pull function. 
     """
    # Required arguments:  pipeline_name 
    #Optional arguments: jsonfilepath, filepath(file path to store mlmd file) 
    output = _metadata_export(pipeline_name, jsonfilepath, filepath)
    return output

def artifact_pull(pipeline_name: str, filepath = "./mlmd"):
    """ Pulls artifacts from the initialized repository.

    Example:
    ```python
         result = artifact_pull("example_pipeline", "./mlmd_directory")
    ```

    Args:
        pipeline_name: Name of the pipeline.
        filepath: Path to store artifacts.
    Returns:
        Output from the _artifact_pull function.
    """

    # Required arguments: Pipeline_name
    # Optional arguments: filepath( path to store artifacts)
    output = _artifact_pull(pipeline_name, filepath)
    return output

def artifact_pull_single(pipeline_name: str, filepath: str, artifact_name: str):
    """ Pulls a single artifact from the initialized repository. 
    Example: 
    ```python 
        result = artifact_pull_single("example_pipeline", "./mlmd_directory", "example_artifact") 
    ```
    Args: 
       pipeline_name: Name of the pipeline. 
       filepath: Path to store the artifact. 
       artifact_name: Name of the artifact. 
    Returns:
       Output from the _artifact_pull_single function. 
    """

    # Required arguments: Pipeline_name
    # Optional arguments: filepath( path to store artifacts), artifact_name
    output = _artifact_pull_single(pipeline_name, filepath, artifact_name)
    return output

def artifact_push(pipeline_name: str, filepath = "./mlmd"):
    """ Pushes artifacts to the initialized repository.

    Example:
    ```python
         result = artifact_push("example_pipeline", "./mlmd_directory")
    ```
    Args: 
       pipeline_name: Name of the pipeline. 
       filepath: Path to store the artifact. 
    Returns:
        Output from the _artifact_push function.
    """

    output = _artifact_push(pipeline_name, filepath)
    return output

def cmf_init_show():
    """ Initializes and shows details of the CMF command. 
    Example: 
    ```python 
         result = cmf_init_show() 
    ``` 
    Returns: 
       Output from the _cmf_cmd_init function. 
    """

    output=_cmf_cmd_init()
    return output

def cmf_init(type: str = "",
        path: str = "",
        git_remote_url: str = "",
        cmf_server_url: str = "",
        neo4j_user: str = "",
        neo4j_password: str = "",
        neo4j_uri: str = "",
        url: str = "",
        endpoint_url: str = "",
        access_key_id: str = "",
        secret_key: str = "",
        session_token: str = "",
        user: str = "",
        password: str = "",
        port: int = 0,
        osdf_path: str = "",
        key_id: str = "",
        key_path: str = "",
        key_issuer: str = "",
         ):

    """ Initializes the CMF configuration based on the provided parameters. 
    Example:
    ```python
       cmf_init( type="local", 
                 path="/path/to/re",
                 git_remote_url="git@github.com:user/repo.git",
                 cmf_server_url="http://cmf-server"
                 neo4j_user", 
                 neo4j_password="password",
                 neo4j_uri="bolt://localhost:76"
               )
    ```
    Args: 
       type: Type of repository ("local", "minioS3", "amazonS3", "sshremote")
       path: Path for the local repository. 
       git_remote_url: Git remote URL for version control.
       cmf_server_url: CMF server URL.
       neo4j_user: Neo4j database username.
       neo4j_password: Neo4j database password.
       neo4j_uri: Neo4j database URI.
       url: URL for MinioS3 or AmazonS3.
       endpoint_url: Endpoint URL for MinioS3.
       access_key_id: Access key ID for MinioS3 or AmazonS3.
       secret_key: Secret key for MinioS3 or AmazonS3. 
       session_token: Session token for AmazonS3.
       user: SSH remote username.
       password: SSH remote password. 
       port: SSH remote port
    Returns:
       Output based on the initialized repository type.
    """

    if type == "":
        return print("Error: Type is not provided")
    if type not in ["local","minioS3","amazonS3","sshremote","osdfremote"]:
        return print("Error: Type value is undefined"+ " "+type+".Expected: "+",".join(["local","minioS3","amazonS3","sshremote","osdfremote"]))

    if neo4j_user != "" and  neo4j_password != "" and neo4j_uri != "":
        pass
    elif neo4j_user == "" and  neo4j_password == "" and neo4j_uri == "":
        pass
    else:
        return print("Error: Enter all neo4j parameters.") 

    args={'path': path,
        'git_remote_url': git_remote_url,
        'url': url,
        'endpoint_url': endpoint_url,
        'access_key_id': access_key_id,
        'secret_key': secret_key,
        'session_token': session_token,
        'user': user,
        'password': password,
        'osdf_path': osdf_path,
        'key_id': key_id,
        'key_path': key_path, 
        'key-issuer': key_issuer,
        }

    status_args=non_related_args(type, args)

    if type == "local" and path != "" and  git_remote_url != "" :
        """Initialize local repository"""
        output = _init_local(
            path, git_remote_url, cmf_server_url, neo4j_user, neo4j_password, neo4j_uri
        )
        if status_args != []:
            print("There are non-related arguments: "+",".join(status_args)+".Please remove them.")
        return output
         
    elif type == "minioS3" and url != "" and endpoint_url != "" and access_key_id != "" and secret_key != "" and git_remote_url != "":
        """Initialize minioS3 repository"""
        output = _init_minioS3(
            url,
            endpoint_url,
            access_key_id,
            secret_key,
            git_remote_url,
            cmf_server_url,
            neo4j_user,
            neo4j_password,
            neo4j_uri,
        )
        if status_args != []:
            print("There are non-related arguments: "+",".join(status_args)+".Please remove them.")
        return output

    elif type == "amazonS3" and url != "" and access_key_id != "" and secret_key != "" and git_remote_url != "":
        """Initialize amazonS3 repository"""
        output = _init_amazonS3(
            url,
            access_key_id,
            secret_key,
            session_token,
            git_remote_url,
            cmf_server_url,
            neo4j_user,
            neo4j_password,
            neo4j_uri,
        )
        if status_args != []:
            print("There are non-related arguments: "+",".join(status_args)+".Please remove them.")

        return output

    elif type == "sshremote" and path != "" and user != "" and port != 0 and password != "" and git_remote_url != "":
        """Initialize sshremote repository"""
        output = _init_sshremote(
            path,
            user,
            port,
            password,
            git_remote_url,
            cmf_server_url,
            neo4j_user,
            neo4j_password,
            neo4j_uri,
        )
        if status_args != []:
            print("There are non-related arguments: "+",".join(status_args)+".Please remove them.")

        return output

    elif type == "osdfremote" and osdf_path != "" and key_id != "" and key_path != 0 and key_issuer != "" and git_remote_url != "":
        """Initialize osdfremote repository"""
        output = _init_osdfremote(
            osdf_path,
            key_id,
            key_path,
            key_issuer,
            git_remote_url,
            cmf_server_url,
            neo4j_user,
            neo4j_password,
            neo4j_uri,
        )
        if status_args != []:
            print("There are non-related arguments: "+",".join(status_args)+".Please remove them.")

        return output

    else:
        print("Error: Enter all arguments")


def non_related_args(type : str, args : dict):
    available_args=[i for i, j in args.items() if j != ""]
    local=["path", "git_remote_url"]
    minioS3=["url", "endpoint_url", "access_key_id", "secret_key", "git_remote_url"]
    amazonS3=["url", "access_key_id", "secret_key", "session_token", "git_remote_url"]
    sshremote=["path", "user", "port", "password", "git_remote_url"]
    osdfremote=["osdf_path", "key_id", "key_path", "key-issuer", "git_remote_url"]


    dict_repository_args={"local" : local, "minioS3" : minioS3, "amazonS3" : amazonS3, "sshremote" : sshremote}
    
    for repo,arg in dict_repository_args.items():
        if repo ==type:
            non_related_args=list(set(available_args)-set(dict_repository_args[repo]))
    return non_related_args




###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

CONTEXT_LIST = ["create_time_since_epoch", "custom_properties", "id", \
                 "last_update_time_since_epoch", "name", "properties", "type", "type_id", "uri", "name", "Execution_uuid"]



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import json
from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cli.utils import find_root
from cmflib.server_interface import server_interface
from cmflib.utils.cmf_config import CmfConfig


# This class pushes mlmd file to cmf-server
class CmdMetadataPush(CmdBase):
    def run(self):
        current_directory = os.getcwd()
        mlmd_file_name = "./mlmd"

        # checks if mlmd filepath is given
        if self.args.file_name:
            mlmd_file_name = self.args.file_name
            current_directory = os.path.dirname(self.args.file_name)

        # checks if mlmd file is present in current directory or given directory
        if not os.path.exists(mlmd_file_name):
            return f"ERROR: {mlmd_file_name} doesn't exists in the {current_directory}."

        query = cmfquery.CmfQuery(mlmd_file_name)
        # print(json.dumps(json.loads(json_payload), indent=4, sort_keys=True))
        execution_flag = 0
        status_code = 0

        # Get url from config
        cmfconfig = os.environ.get("CONFIG_FILE",".cmfconfig")

        # find root_dir of .cmfconfig
        output = find_root(cmfconfig)

        # in case, there is no .cmfconfig file
        if output.find("'cmf' is  not configured") != -1:
            return output

        config_file_path = os.path.join(output, cmfconfig)
        attr_dict = CmfConfig.read_config(config_file_path)
        url = attr_dict.get("cmf-server-ip", "http://127.0.0.1:80")

        print("metadata push started")
        print("........................................")

        # Checks if pipeline name exists
        if self.args.pipeline_name in query.get_pipeline_names():
            # converts mlmd file to json format
            json_payload = query.dumptojson(self.args.pipeline_name, None)
            # checks if execution_id is given by user
            if self.args.execution:
                exec_id = self.args.execution
                mlmd_data = json.loads(json_payload)["Pipeline"]
                # checks if given execution_id present in mlmd
                for i in mlmd_data[0]["stages"]:
                    for j in i["executions"]:
                        if j["id"] == int(exec_id):
                            execution_flag = 1
                            # calling mlmd_push api to push mlmd file to cmf-server
                            response = server_interface.call_mlmd_push(
                                json_payload, url, exec_id, self.args.pipeline_name
                            )
                            break
                if execution_flag == 0:
                    return "Given execution is not found in mlmd."
            else:
                exec_id = None
                response = server_interface.call_mlmd_push(json_payload, url, exec_id, self.args.pipeline_name)
            status_code = response.status_code
            if status_code == 200 and response.json()['status']=="success":
                print("mlmd is successfully pushed.")
            elif status_code==200 and response.json()["status"]=="exists":
                print("Executions already exists.")
            elif status_code==422 and response.json()["status"]=="version_update":
                return "ERROR: You need to update cmf to the latest version. Unable to push metadata file."
            elif status_code == 404:
                return "ERROR: cmf-server is not available."
            elif status_code == 500:
                return "ERROR: Internal server error."
            else:
                return "ERROR: Status Code = {status_code}. Unable to push mlmd."

            if self.args.tensorboard:
                # /tensorboard api call is done only if mlmd push is successfully completed
                # tensorboard parameter is passed
                print("......................................")
                print("tensorboard logs upload started!!")
                print("......................................")

                # check if the path provided is for a file
                if os.path.isfile(self.args.tensorboard):
                    file_name = os.path.basename(self.args.tensorboard)
                    tresponse = server_interface.call_tensorboard(url, self.args.pipeline_name, file_name, self.args.tensorboard)
                    tstatus_code = tresponse.status_code
                    if tstatus_code == 200:
                        return "tensorboard logs: file {file_name} pushed successfully"
                    else:
                        return "ERROR: Failed to upload file {file_name}. Server response: {response.text}"
                # If path provided is a directory
                elif os.path.isdir(self.args.tensorboard):
                    # Recursively push all files and subdirectories
                    for root, dirs, files in os.walk(self.args.tensorboard):
                        for file_name in files:
                            file_path = os.path.join(root, file_name)
                            relative_path = os.path.relpath(file_path, self.args.tensorboard)
                            tresponse = server_interface.call_tensorboard(url, self.args.pipeline_name, relative_path, file_path)
                            if tresponse.status_code == 200:
                                print(f"tensorboard logs: File {file_name} uploaded successfully.")
                            else:
                                return f"ERROR: Failed to upload file {file_name}. Server response: {tresponse.text}"
                    return f"tensorboard logs: {self.args.tensorboard} uploaded successfully!!"
                else:
                    return "ERROR: Invalid data path. Provide valid file/folder path for tensorboard logs!!"
            else:
                return "SUCCESS!!"
        else:
            return "Pipeline name " + self.args.pipeline_name + " doesn't exists."


def add_parser(subparsers, parent_parser):
    PUSH_HELP = "Push user-generated mlmd to server to create one single mlmd file for all the pipelines."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push user's mlmd to cmf-server.",
        help=PUSH_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", "--file_name", help="Specify mlmd file name.", metavar="<file_name>"
    )

    parser.add_argument(
        "-e",
        "--execution",
        help="Specify Execution id.",
        metavar="<exec_id>",
    )

    parser.add_argument(
        "-t",
        "--tensorboard",
        help="Specify path to tensorboard logs for the pipeline.",
        metavar="<tensorboard>"
    )

    parser.set_defaults(func=CmdMetadataPush)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


# Abstract class for commands
class CmdBase(ABC):
    def __init__(self, args):
        self.args = args

    def do_run(self):
        return self.run()

    @abstractmethod
    def run(self):
        pass



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
from dvc.api import DVCFileSystem

class LocalArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        obj = True
        try:
            fs = DVCFileSystem(
                dvc_config_op["remote.local-storage.url"]
            )  # dvc_config_op[1] is file system path - "/path/to/local/repository"
            
            # get_file() only creates file, to put artifacts in proper directory, subfolders are required.
            # download_loc = contains absolute path of the file with file name and extension
            dir_path = ""
            if "/" in download_loc:
                dir_path, _ = download_loc.rsplit("/", 1)
            if dir_path != "":
                os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders if needed

            response = ""

            """"
            if object_name ends with .dir - it is a directory.
            we download .dir object with 'temp_dir' and remove 
            this after all the files from this .dir object is downloaded.
            """
            if object_name.endswith('.dir'):
                # in case of .dir, download_loc is a absolute path for a folder
                os.makedirs(download_loc, mode=0o777, exist_ok=True)

                # download the .dir object 
                temp_dir = f"{download_loc}/dir"
                response = fs.get_file(object_name, temp_dir)

                with open(temp_dir, 'r') as file:
                    tracked_files = eval(file.read())

                # removing temp_dir
                if os.path.exists(temp_dir):
                    os.remove(temp_dir)

                """
                object_name = "files/md5/9b/9a458ac0b534f088a47c2b68bae479.dir" 
                contains the path of the .dir on the artifact repo
                we need to remove the hash of the .dir from the object_name
                which will leave us with the artifact repo path
                """
                repo_path = "/".join(object_name.split("/")[:-2])
                for file_info in tracked_files:
                    relpath = file_info['relpath']
                    md5_val = file_info['md5']
                    # md5_val = a237457aa730c396e5acdbc5a64c8453
                    # we need a2/37457aa730c396e5acdbc5a64c8453
                    formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                    temp_object_name = f"{repo_path}/{formatted_md5}"
                    temp_download_loc = f"{download_loc}/{relpath}"
                    obj = fs.get_file(temp_object_name, temp_download_loc)
                    if obj == None: 
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
            else:
                response = fs.get_file(object_name, download_loc)
            if response == None:  # get_file() returns none when file gets downloaded.
                stmt = f"object {object_name} downloaded at {download_loc}."
                return stmt
        except TypeError as exception:
            return exception
        except Exception as exception:
            return exception



from ray import tune
from ray.tune import Callback
from cmflib import cmf

class CmfRayLogger(Callback):
    #id_count = 1
    
    def __init__(self, pipeline_name, file_path, pipeline_stage):
        """
        pipeline_name: The name of the CMF Pipelibe
        file_path: The path to metadata file
        pipeline_stage: The name for the stage of cmf_pipeline
        """
        self.pipeline_name = pipeline_name
        self.file_path = file_path
        self.pipeline_stage = pipeline_stage
        self.cmf_obj = {}

    def on_trial_start(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        print(f"CMF Logging Started for Trial {trial_id}")
        self.cmf_obj[trial_id] = cmf.Cmf(filepath = self.file_path, pipeline_name = self.pipeline_name)
        _ = self.cmf_obj[trial_id].create_context(pipeline_stage = self.pipeline_stage)
        _ = self.cmf_obj[trial_id].create_execution(execution_type=f"Trial_{trial_id}",
                                            create_new_execution = False)
        #self.execution_id[trial_id] = CmfRayLogger.id_count
        #CmfRayLogger.id_count+=1

    def on_trial_result(self, iteration, trials, trial, result, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result
        curr_res = result
        print(f'In Trial Results')
        print(f'Trial_ID: {trial_id}')
        print(f'curr_results is {curr_res}')
        print(f'Logging results to CMF with metric name Trial_{trial_id}_metrics')
        #if trial_id in self.execution_id:
        #    _ = self.metawriter.update_execution(int(self.execution_id[trial_id]))
        _ = self.cmf_obj[trial_id].log_metric(metrics_name = f"Trial_{trial_id}_metrics",
                                      custom_properties = {'Output': curr_res})
        
        
    def on_trial_complete(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result
        
        print(f"Trial {trial_id} completed, Commiting to CMF: with name Trial_{trial_id}_metrics")
        print()
        #if trial_id in self.execution_id:
        #    _ = self.metawriter.update_execution(int(self.execution_id[trial_id]))
        _ = self.cmf_obj[trial_id].commit_metrics(f"Trial_{trial_id}_metrics")
        
    def on_trial_error(self, iteration, trials, trial, **info):
        trial_id = trial.trial_id
        trial_config = trial.config
        trial_result = trial.last_result

        print(f"An error occured with Trial {trial_id}, Not commiting anything to cmf")


import itertools
import re
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout
import pandas as pd
from cmflib import cmfquery
import dvc
import json
import random
import warnings

warnings.filterwarnings("ignore")
def query_list_of_executions(mlmd_path, pipeline_name, dict_of_art_ids, dict_of_exe_ids):
    list_of_exec = []
    list_of_exec_uuid = []
    list_of_exec = dict_of_exe_ids[pipeline_name]["Context_Type"].tolist()
    list_of_uuid = dict_of_exe_ids[pipeline_name]["Execution_uuid"].tolist()
    for exec_type, uuid in zip(list_of_exec, list_of_uuid):
        list_of_exec_uuid.append(exec_type.split("/",1)[1] + "_" + uuid.split("-")[0][:4])
    return list_of_exec_uuid

#print(query_visualization_execution("/home/chobey/cmf-server/data/mlmd","image"))



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import subprocess

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cli.utils import find_root
from cmflib.dvc_wrapper import dvc_get_config
from cmflib.utils.cmf_config import CmfConfig

class CmdInitShow(CmdBase):
    def run(self):
        cmfconfig = os.environ.get("CONFIG_FILE",".cmfconfig")
        msg = "'cmf' is not configured.\nExecute 'cmf init' command."
        result = dvc_get_config()
        if len(result) == 0:
            return msg
        else:
            cmf_config_root = find_root(cmfconfig)
            if cmf_config_root.find("'cmf' is not configured") != -1:
                return msg
            config_file_path = os.path.join(cmf_config_root, cmfconfig)
            attr_dict = CmfConfig.read_config(config_file_path)
            attr_list = []
            for key, value in attr_dict.items():
                temp_str = f"{key} = {value}"
                attr_list.append(temp_str)
            attr_str = "\n".join(attr_list)
            return f"{result}\n{attr_str}"


def add_parser(subparsers, parent_parser):
    HELP = "Show current CMF configuration."

    parser = subparsers.add_parser(
        "show",
        parents=[parent_parser],
        description="This command shows current cmf configuration including cmf server ip.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(func=CmdInitShow)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import argparse

from cmflib.commands.artifact import pull, push
from cmflib.cli.utils import *

SUB_COMMANDS = [pull, push]

# This parser adds positional arguments to the main parser
def add_parser(subparsers, parent_parser):
    ARTIFACT_HELP = "Command for artifact pull/push."

    artifact_parser = subparsers.add_parser(
        "artifact",
        parents=[parent_parser],
        description="Pull or Push artifacts as per current cmf configuration.",
        help=ARTIFACT_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    artifact_subparsers = artifact_parser.add_subparsers(
        dest="cmd", help="Use `cmf artifact CMD --help` for " "command-specific help."
    )

    fix_subparsers(artifact_subparsers)
    for cmd in SUB_COMMANDS:
        cmd.add_parser(artifact_subparsers, parent_parser)
